"""
Application involving approach trajectory of a bottle opener towards a bottle
"""

# Imports
import time as t
import numpy as np
from mpl_toolkits.mplot3d import Axes3D
from invariants_py.rockit_calculate_vector_invariants_position import OCP_calc_pos
from invariants_py.ocp_helper import solution_check_pos
from invariants_py.rockit_generate_position_from_vector_invariants import OCP_gen_pos
import invariants_py.data_handler as dh
import matplotlib.pyplot as plt
import invariants_py.reparameterization as reparam
import scipy.interpolate as ip
import invariants_py.plotters as pl

t0 = t.time()
show_plots = False
use_fatrop_solver = True  # True = fatrop, False = ipopt
ocp_to_func = True

# Read trajectory data
data_location = dh.find_data_path('beer_1.txt')
trajectory, time = dh.read_pose_trajectory_from_txt(data_location)
pose, time_profile, arclength, nb_samples, stepsize = reparam.reparameterize_trajectory_arclength(trajectory)
arclength_n = arclength / arclength[-1]
trajectory = pose[:, 0:3, 3]

if show_plots:
    fig = plt.figure(figsize=(8, 8))
    ax = fig.add_subplot(111, projection='3d')
    ax = plt.axes(projection='3d')
    ax.plot(trajectory[:, 0], trajectory[:, 1], trajectory[:, 2], '.-')

# Calculate invariants given measurements
rms_error_traj = 0.004
FS_calculation_problem = OCP_calc_pos(window_len=nb_samples, bool_unsigned_invariants=False, rms_error_traj=rms_error_traj, fatrop_solver=use_fatrop_solver)
invariants, calculate_trajectory, movingframes = FS_calculation_problem.calculate_invariants(trajectory, stepsize)
solution_check_pos(trajectory, calculate_trajectory, rms_error_traj)
init_vals_calculate_trajectory = calculate_trajectory
init_vals_movingframes = movingframes

if show_plots:
    fig = plt.figure(figsize=(8, 8))
    ax = fig.add_subplot(111, projection='3d')
    ax = plt.axes(projection='3d')
    ax.plot(trajectory[:, 0], trajectory[:, 1], trajectory[:, 2], '.-')
    ax.plot(calculate_trajectory[:, 0], calculate_trajectory[:, 1], calculate_trajectory[:, 2], '.-')
    pl.plot_invariants(invariants, [], arclength_n, [], inv_type='FS_pos')
    if plt.get_backend() != 'agg':
        plt.show()

# Spline of model
knots = np.concatenate(([arclength_n[0]], [arclength_n[0]], arclength_n, [arclength_n[-1]], [arclength_n[-1]]))
degree = 3
spline_model_trajectory = ip.BSpline(knots, invariants, degree)


def interpolate_model_invariants(demo_invariants, progress_values):
    resampled_invariants = np.array([demo_invariants(i) for i in progress_values])
    new_stepsize = progress_values[1] - progress_values[0]
    resampled_invariants[:, 0] = resampled_invariants[:, 0] * (progress_values[-1] - progress_values[0])
    return resampled_invariants, new_stepsize


current_progress = 0
number_samples = 100
calculate_trajectory = init_vals_calculate_trajectory
movingframes = init_vals_movingframes
progress_values = np.linspace(current_progress, arclength_n[-1], number_samples)
model_invariants, new_stepsize = interpolate_model_invariants(spline_model_trajectory, progress_values)

if show_plots:
    pl.plot_interpolated_invariants(invariants, model_invariants, arclength_n, progress_values, inv_type='FS_pos')

# New constraints
current_index = round(current_progress * len(trajectory))
boundary_constraints = {
    "position": {
        "initial": calculate_trajectory[current_index],
        "final": calculate_trajectory[-1] + np.array([0.1, 0.05, 0.05])
    },
    "moving-frame": {
        "initial": movingframes[current_index],
        "final": movingframes[-1]
    },
}
initial_values = {
    "trajectory": calculate_trajectory,
    "moving-frames": movingframes,
    "invariants": model_invariants,
}

# Generate new trajectory
FS_online_generation_problem = OCP_gen_pos(boundary_constraints, window_len=number_samples, fatrop_solver=use_fatrop_solver)
weights = {}
weights['w_invars'] = np.array([5 * 10 ** 1, 1.0, 1.0])

if ocp_to_func:
    new_invars, new_trajectory, new_movingframes, tot_time_pos = FS_online_generation_problem.generate_trajectory(
        invariant_model=model_invariants, initial_values=initial_values, boundary_constraints=boundary_constraints,
        step_size=new_stepsize, weights_params=weights)
else: 
    new_invars, new_trajectory, new_movingframes, tot_time_pos = FS_online_generation_problem.generate_trajectory_OLD(
    invariant_model=model_invariants, initial_values=initial_values, boundary_constraints=boundary_constraints,
    step_size=new_stepsize, weights_params=weights)

# print(new_invars)
# input("Press Enter to continue...")

if use_fatrop_solver:
    print('')
    print("TOTAL time to generate new trajectory: ")
    print(str(tot_time_pos) + '[s]')

if show_plots:
    fig = plt.figure(figsize=(14, 8))
    ax = fig.add_subplot(111, projection='3d')
    ax.plot(calculate_trajectory[:, 0], calculate_trajectory[:, 1], calculate_trajectory[:, 2], 'b')
    ax.plot(new_trajectory[:, 0], new_trajectory[:, 1], new_trajectory[:, 2], 'r')
    ax.set_xlabel('x')
    ax.set_ylabel('y')
    ax.set_zlabel('z')
    pl.plot_invariants(invariants, new_invars, arclength_n, progress_values, inv_type='FS_pos')
    if plt.get_backend() != 'agg':
        plt.show()

# Visualization
window_len = 20
FS_online_generation_problem2 = OCP_gen_pos(boundary_constraints, window_len=window_len, fatrop_solver=use_fatrop_solver)
current_progress = 0.0
old_progress = 0.0
calculate_trajectory = init_vals_calculate_trajectory
movingframes = init_vals_movingframes

plt.ion()  # Interactive plotting
fig_traj = plt.figure(figsize=(8, 8))
fig_invars = plt.figure(figsize=(12, 4))

while current_progress <= 1.0:
    print(f"current progress = {current_progress}")

    # Resample invariants for current progress
    progress_values = np.linspace(current_progress, arclength_n[-1], window_len)
    model_invariants, new_stepsize = interpolate_model_invariants(spline_model_trajectory, progress_values)

    # Boundary constraints
    current_index = round((current_progress - old_progress) * len(calculate_trajectory))
    #boundary_constraints["position"]["initial"] = calculate_trajectory[current_index]
    boundary_constraints = {
        "position": {
            "initial": calculate_trajectory[current_index],
            "final": trajectory[-1] - current_progress * np.array([-0.2, 0.0, 0.0])
        },
        "moving-frame": {
            "initial": movingframes[current_index],
            "final": movingframes[-1]
        },
        #  speed/acceleration values - velocity/(acceleration vectors) - moving frame
    }
    initial_values = {
        "trajectory": calculate_trajectory,
        "moving-frames": movingframes,
        "invariants": model_invariants,
    }

    # Calculate remaining trajectory
    if ocp_to_func:
        new_invars, calculate_trajectory, movingframes, tot_time_pos = FS_online_generation_problem2.generate_trajectory(
        invariant_model=model_invariants, initial_values=initial_values, boundary_constraints=boundary_constraints,
        step_size=new_stepsize, weights_params=weights)
    else:
        new_invars, calculate_trajectory, movingframes, tot_time_pos = FS_online_generation_problem2.generate_trajectory_OLD(
    invariant_model=model_invariants, initial_values=initial_values, boundary_constraints=boundary_constraints,
    step_size=new_stepsize, weights_params=weights)

    if use_fatrop_solver:
        print('')
        print("TOTAL time to generate new trajectory: ")
        print(str(tot_time_pos) + '[s]')

    if show_plots:
        fig_traj.clf()
        ax_traj = fig_traj.add_subplot(111, projection='3d')
        ax_traj.view_init(elev=26, azim=-40)
        ax_traj.plot(trajectory[:, 0], trajectory[:, 1], trajectory[:, 2], 'b')
        ax_traj.plot(calculate_trajectory[:, 0], calculate_trajectory[:, 1], calculate_trajectory[:, 2], 'r')
        fig_traj.canvas.draw()
        fig_traj.canvas.flush_events()

    # print(new_invars)
    # input("Press Enter to continue...")

    if show_plots:
        fig_invars.clf()
        ax_inv = fig_invars.add_subplot(131)
        ax_inv.plot(progress_values, new_invars[:, 0], 'r')
        ax_inv.plot(arclength_n, invariants[:, 0], 'b')
        ax_inv.set_title('velocity [m/m]')
        ax_inv = fig_invars.add_subplot(132)
        ax_inv.plot(progress_values, new_invars[:, 1], 'r')
        ax_inv.plot(arclength_n, invariants[:, 1], 'b')
        ax_inv.set_title('curvature [rad/m]')
        ax_inv = fig_invars.add_subplot(133)
        ax_inv.plot(progress_values, new_invars[:, 2], 'r')
        ax_inv.plot(arclength_n, invariants[:, 2], 'b')
        ax_inv.set_title('torsion [rad/m]')
        fig_invars.canvas.draw()
        fig_invars.canvas.flush_events()

    old_progress = current_progress
    current_progress = old_progress + 1 / window_len

if show_plots:
    t.sleep(3)
    plt.close(fig_invars)
    plt.close(fig_traj)

t1 = t.time()
print(f"Total time: {t1 - t0} [s]")

# print(new_invars)
# input("Press Enter to continue...")


import numpy as np
import invariants_py.SO3 as SO3
from invariants_py.reparameterization import interpR

def FSr_init(R_obj_start,R_obj_end,N=100):
    skew_angle = SO3.logm(R_obj_start.T @ R_obj_end)
    angle_vec_in_body = np.array([skew_angle[2,1],skew_angle[0,2],skew_angle[1,0]])
    angle_vec_in_world = R_obj_start@angle_vec_in_body
    angle_norm = np.linalg.norm(angle_vec_in_world)
    U_init = np.tile(np.array([angle_norm,0.001,0.001]),(N-1,1))
    e_x_fs_init = angle_vec_in_world/angle_norm
    e_y_fs_init = [0,1,0]
    e_y_fs_init = e_y_fs_init - np.dot(e_y_fs_init,e_x_fs_init)*e_x_fs_init
    e_y_fs_init = e_y_fs_init/np.linalg.norm(e_y_fs_init)
    e_z_fs_init = np.cross(e_x_fs_init,e_y_fs_init)
    R_r_init = np.array([e_x_fs_init,e_y_fs_init,e_z_fs_init]).T

    R_r_init_array = []
    for k in range(N):
        R_r_init_array.append(R_r_init)
    R_r_init_array = np.array(R_r_init_array)

    return R_r_init, R_r_init_array, U_init

def generate_initvals_from_bounds(boundary_constraints,N):
    
    # Generate initial trajectory using linear interpolation
    p1 = boundary_constraints["position"]["final"]
    p0 = boundary_constraints["position"]["initial"]
    initial_trajectory = np.linspace(p0, p1, N).T

    # Generate corresponding initial invariants
    diff_vector = np.array(p1) - np.array(p0)
    L = np.linalg.norm(diff_vector)
    initial_invariants = np.tile(np.array([[L],[0.0001],[0.0001]]),(1,N-1))

    # Generate corresponding initial moving frames using Gram-Schmidt process
    e_x = diff_vector / L
    e_y = np.array([0, 1, 0]) - np.dot(np.array([0, 1, 0]), e_x) * e_x
    e_y = e_y / np.linalg.norm(e_y)
    e_z = np.cross(e_x, e_y)
    R_mf = np.column_stack((e_x, e_y, e_z))
    initial_movingframes = np.tile(R_mf, (N,1,1))

    initial_values = {
         "trajectory": initial_trajectory.T,
         "moving-frames": initial_movingframes,
         "invariants": initial_invariants
    }

    R_t_x_sol = np.tile(e_x, (N, 1)).T
    R_t_y_sol = np.tile(e_y, (N, 1)).T
    R_t_z_sol = np.tile(e_z, (N, 1)).T

    return [initial_invariants, initial_trajectory, R_t_x_sol, R_t_y_sol, R_t_z_sol], initial_values

def generate_initvals_from_bounds_rot(boundary_constraints,N):
    R0 = boundary_constraints["orientation"]["initial"]
    R1 = boundary_constraints["orientation"]["final"]
    # Linear initialization
    initial_trajectory = interpR(np.linspace(0, 1, N), [0,1], np.array([R0, R1]))

    _, R_r_sol, initial_invariants = FSr_init(R0, R1)
    R_r_sol_x = R_r_sol[:,:,0].T
    R_r_sol_y = R_r_sol[:,:,1].T
    R_r_sol_z = R_r_sol[:,:,2].T

    R_obj_sol_x = initial_trajectory[:,:,0].T
    R_obj_sol_y = initial_trajectory[:,:,1].T
    R_obj_sol_z = initial_trajectory[:,:,2].T


    return [initial_invariants.T, R_r_sol_x, R_r_sol_y, R_r_sol_z, R_obj_sol_x, R_obj_sol_y, R_obj_sol_z]

def calculate_velocity_from_discrete_rotations(R, timestamps):
    """
    In this code, rotational velocity is calculated based on discrete
    rotations using central differences.

    Input:
        R: rotation matrix                      (Nx3x3)
        timestamps: time                        (Nx1)
    Output:
        rot_velocity: rotational velocity       (Nx3)
    """
    
    N = R.shape[0]
    rot_velocity = np.zeros((N, 3))

    # First sample
    DeltaR = R[1,:,:] @ R[0,:,:].T
    dt = timestamps[1] - timestamps[0]
    dtwist = SO3.logm(DeltaR) / dt
    omega = np.array([-dtwist[1, 2], dtwist[0, 2], -dtwist[0, 1]])
    rot_velocity[0, :] = omega

    # Middle samples (central difference)
    for i in range(1, N - 1):
        DeltaR = R[i+1,:,:] @ R[i-1,:,:].T
        dt = timestamps[i + 1] - timestamps[i - 1]
        dtwist = SO3.logm(DeltaR) / dt
        omega = np.array([-dtwist[1, 2], dtwist[0, 2], -dtwist[0, 1]])
        rot_velocity[i, :] = omega

    # Last sample
    DeltaR = R[N-1,:,:] @ R[N-2,:,:].T
    dt = timestamps[N - 1] - timestamps[N - 2]
    dtwist = SO3.logm(DeltaR) / dt
    omega = np.array([-dtwist[1, 2], dtwist[0, 2], -dtwist[0, 1]])
    rot_velocity[N - 1, :] = omega

    return rot_velocity

def estimate_initial_frames(vector_traj):    
    # Estimate initial moving frames based on measurements
    
    N = np.size(vector_traj,0)
    
    #TODO  this is not correct yet, ex not perpendicular to ey + not robust for singularities, these parts must still be transferred from Matlab
    
    ex = vector_traj / np.linalg.norm(vector_traj,axis=1).reshape(N,1)
    ez = np.tile( np.array((0,0,1)), (N,1) )
    ey = np.array([ np.cross(ez[i,:],ex[i,:]) for i in range(N) ])

    return ex,ey,ez

def initialize_VI_pos(input_trajectory):

    if input_trajectory.shape[1] == 3:
        measured_positions = input_trajectory
    else:
        measured_positions = input_trajectory[:,:3,3]

    N = np.size(measured_positions,0)
    Pdiff = np.diff(measured_positions, axis=0)
    Pdiff = np.vstack((Pdiff, Pdiff[-1]))

    [ex,ey,ez] = estimate_initial_frames(Pdiff)

    R_t = np.zeros((3,3*N))
    for i in range(N-1):
        R_t[:,3*i:3*(i+1)] = np.array([ex[i,:],ey[i,:],ez[i,:]])   

    p_obj_sol =  measured_positions.T 
    invars = np.vstack((1e0*np.ones((1,N-1)),1e-1*np.ones((1,N-1)), 1e-12*np.ones((1,N-1))))
    return [invars, p_obj_sol, R_t], measured_positions

def initialize_VI_rot(input_trajectory):

    if input_trajectory.shape[1] == 3:
        measured_orientation = input_trajectory
    else:
        measured_orientation = input_trajectory[:,:3,:3]

    N = np.size(measured_orientation,0)
    Rdiff = calculate_velocity_from_discrete_rotations(measured_orientation,timestamps=np.arange(N))

    [ex,ey,ez] = estimate_initial_frames(Rdiff)

    R_r_x_sol =  ex.T 
    R_r_y_sol =  ey.T 
    R_r_z_sol =  ez.T 
    R_obj_x =  measured_orientation[:,:,0].T
    R_obj_y =  measured_orientation[:,:,1].T
    R_obj_z =  measured_orientation[:,:,2].T

    invars = np.vstack((1e0*np.ones((1,N-1)),1e-1*np.ones((1,N-1)), 1e-12*np.ones((1,N-1))))

    return [invars, R_obj_x, R_obj_y, R_obj_z, R_r_x_sol, R_r_y_sol, R_r_z_sol], measured_orientation


"""
Application involving approach trajectory of a bottle opener towards a bottle
"""

# Imports
import numpy as np
from math import pi
import os
import time as t
import invariants_py.data_handler as dh
import matplotlib.pyplot as plt
import invariants_py.reparameterization as reparam
import scipy.interpolate as ip
from invariants_py.rockit_calculate_vector_invariants_rotation import OCP_calc_rot as FS_calc
from invariants_py.rockit_generate_rotation_from_vector_invariants import OCP_gen_rot as FS_gen
from IPython.display import clear_output
from invariants_py.plotters import plot_3d_frame, plot_orientation, plot_stl
from scipy.spatial.transform import Rotation as R
from invariants_py.orthonormalize_rotation import orthonormalize_rotation as orthonormalize
import invariants_py.plotters as pl
from invariants_py.reparameterization import interpR
from invariants_py.initialization import FSr_init
from invariants_py.ocp_helper import solution_check_rot

#%%
data_location = dh.find_data_path('beer_1.txt')
trajectory,time = dh.read_pose_trajectory_from_txt(data_location)
pose,time_profile,arclength,nb_samples,stepsize = reparam.reparameterize_trajectory_arclength(trajectory)
arclength_n = arclength/arclength[-1]
trajectory_position = pose[:,:3,3]
trajectory_orientation = pose[:,:3,:3]

fig = plt.figure(figsize=(8,8))
ax = fig.add_subplot(111, projection='3d')
ax = plt.axes(projection='3d')
ax.plot(trajectory_position[:,0],trajectory_position[:,1],trajectory_position[:,2],'.-')
n_frames = 10
indx = np.trunc(np.linspace(0,len(trajectory_orientation)-1,n_frames))
indx = indx.astype(int)
opener_location = dh.find_data_path('opener.stl')
for i in indx:
    plot_3d_frame(trajectory_position[i,:],trajectory_orientation[i,:,:],1,0.05,['red','green','blue'],ax)
    plot_stl(opener_location,trajectory_position[i,:],trajectory_orientation[i,:,:],colour="c",alpha=0.2,ax=ax)

#%%
use_fatrop_solver = True # True = fatrop, False = ipopt

# specify optimization problem symbolically
rms_error_traj = 2*pi/180
FS_calculation_problem = FS_calc(window_len=nb_samples, bool_unsigned_invariants = False, rms_error_traj = 2*pi/180, fatrop_solver = use_fatrop_solver) 

# calculate invariants given measurements
invariants, calculate_trajectory, movingframes = FS_calculation_problem.calculate_invariants(trajectory,stepsize)
solution_check_rot(trajectory_orientation,calculate_trajectory,rms_error_traj)

init_vals_calculate_trajectory = calculate_trajectory
init_vals_movingframes = movingframes

fig = plt.figure(figsize=(8,8))
ax = fig.add_subplot(111, projection='3d')
ax = plt.axes(projection='3d')
ax.plot(trajectory_position[:,0],trajectory_position[:,1],trajectory_position[:,2],'.-')
indx = np.trunc(np.linspace(0,len(calculate_trajectory)-1,n_frames))
indx = indx.astype(int)
for i in indx:
    plot_3d_frame(trajectory_position[i,:],calculate_trajectory[i,:,:],1,0.05,['red','green','blue'],ax)
    plot_stl(opener_location,trajectory_position[i,:],calculate_trajectory[i,:,:],colour="r",alpha=0.2,ax=ax)
    plot_stl(opener_location,trajectory_position[i,:],trajectory_orientation[i,:,:],colour="c",alpha=0.2,ax=ax)

plot_orientation(trajectory_orientation,init_vals_calculate_trajectory)

pl.plot_invariants(invariants,[],arclength_n,[],inv_type='FS_rot')

if plt.get_backend() != 'agg':
    plt.show()

#%%
# Spline of model
knots = np.concatenate(([arclength_n[0]],[arclength_n[0]],arclength_n,[arclength_n[-1]],[arclength_n[-1]]))
degree = 3
spline_model_trajectory = ip.BSpline(knots,invariants,degree)

def interpolate_model_invariants(demo_invariants, progress_values):
    
    resampled_invariants = np.array([demo_invariants(i) for i in progress_values]) 
    new_stepsize = progress_values[1] - progress_values[0] 
    
    resampled_invariants[:,0] = resampled_invariants[:,0] *  (progress_values[-1] - progress_values[0])
    return resampled_invariants, new_stepsize

#%% 
current_progress = 0
number_samples = 100

progress_values = np.linspace(current_progress, arclength_n[-1], number_samples)
model_invariants,new_stepsize = interpolate_model_invariants(spline_model_trajectory,progress_values)

pl.plot_interpolated_invariants(invariants, model_invariants, arclength_n, progress_values, inv_type = 'FS_rot')

# new constraints
current_index = round(current_progress*len(trajectory))
R_obj_start = orthonormalize(calculate_trajectory[current_index,:,:])
rotate = R.from_euler('z', 30, degrees=True)
R_obj_end =  orthonormalize(rotate.as_matrix() @ calculate_trajectory[-1])
# R_r_start = orthonormalize(movingframes[current_index])
# R_r_end = orthonormalize(movingframes[-1])

# Linear initialization
R_obj_init = interpR(np.linspace(0, 1, len(calculate_trajectory)), [0,1], np.array([R_obj_start, R_obj_end]))

R_r_init, R_r_init_array, invars_init = FSr_init(R_obj_start, R_obj_end)

initial_values = {"invariants-orientation": invars_init, "trajectory-orientation": R_obj_init, "moving-frame-orientation": R_r_init_array}

weight_params = {"w_invars": 10**2*np.array([10**1, 1.0, 1.0])}

boundary_constraints = {"orientation": {"initial": R_obj_start, "final": R_obj_end}, "moving-frame-orientation": {"initial": R_r_init, "final": R_r_init}}

# specify optimization problem symbolically
FS_online_generation_problem = FS_gen(boundary_constraints, number_samples, fatrop_solver = use_fatrop_solver)

# Solve
new_invars, new_trajectory, new_movingframes, tot_time_rot = FS_online_generation_problem.generate_trajectory(model_invariants,boundary_constraints,new_stepsize,weight_params,initial_values)

if use_fatrop_solver:
    print('')
    print("TOTAL time to generate new trajectory: ")
    print(str(tot_time_rot) + '[s]')

for i in range(len(new_trajectory)):
    new_trajectory[i] = orthonormalize(new_trajectory[i])

fig = plt.figure(figsize=(14,8))
ax = fig.add_subplot(111, projection='3d')
ax.plot(trajectory_position[:,0],trajectory_position[:,1],trajectory_position[:,2],'.-')
for i in indx:
    plot_stl(opener_location,trajectory_position[i,:],calculate_trajectory[i,:,:],colour="c",alpha=0.2,ax=ax)

cut_trajectory_position = trajectory_position[current_index:,:]
interp = ip.interp1d(np.linspace(0,1,len(cut_trajectory_position)),cut_trajectory_position, axis = 0)
trajectory_position_online = interp(np.linspace(0,1,len(new_trajectory)))
indx_online = np.trunc(np.linspace(0,len(new_trajectory)-1,n_frames))
indx_online = indx_online.astype(int)
for i in indx_online:
    plot_3d_frame(trajectory_position_online[i,:],new_trajectory[i,:,:],1,0.05,['red','green','blue'],ax)
    plot_stl(opener_location,trajectory_position_online[i,:],new_trajectory[i,:,:],colour="r",alpha=0.2,ax=ax)
plot_orientation(calculate_trajectory, new_trajectory,current_index)

pl.plot_invariants(invariants,new_invars,arclength_n,progress_values,inv_type='FS_rot')

fig99 = plt.figure(figsize=(14,8))
ax99 = fig99.add_subplot(111, projection='3d')
pl.plot_stl(opener_location,[0,0,0],calculate_trajectory[-1],colour="r",alpha=0.5,ax=ax99)
pl.plot_stl(opener_location,[0,0,0],R_obj_end,colour="b",alpha=0.5,ax=ax99)
pl.plot_stl(opener_location,[0,0,0],new_trajectory[-1],colour="g",alpha=0.5,ax=ax99)

if plt.get_backend() != 'agg':
    plt.show()

#%% Visualization

window_len = 20

# specify optimization problem symbolically
FS_online_generation_problem2 = FS_gen(boundary_constraints, window_len, fatrop_solver = use_fatrop_solver)

current_progress = 0.0
old_progress = 0.0

# plt.ion()  # Interactive plotting
# fig_traj = plt.figure(figsize=(8, 8))
# fig_invars = plt.figure(figsize=(12, 4))

R_obj_end = calculate_trajectory[-1]
trajectory_position_iter = trajectory_position.copy()
fig_invars = plt.figure(figsize=(10, 6))
fig = plt.figure(figsize=(14,8))
ax = fig.add_subplot(111, projection='3d')
while current_progress <= 1.0:
    
    print(f"current progress = {current_progress}")

    # Resample invariants for current progress
    progress_values = np.linspace(current_progress, arclength_n[-1], window_len)
    model_invariants,new_stepsize = interpolate_model_invariants(spline_model_trajectory,progress_values)
    
    # Boundary constraints
    current_index = round( (current_progress - old_progress) * len(calculate_trajectory))
    rotate = R.from_euler('z', 30/window_len, degrees=True)
    R_obj_end =  orthonormalize(rotate.apply(R_obj_end))
    boundary_constraints = {"orientation": {"initial": calculate_trajectory[current_index], "final": R_obj_end}, "moving-frame-orientation": {"initial": movingframes[current_index], "final": movingframes[-1]}}

    initial_values = {"invariants-orientation": model_invariants, "trajectory-orientation": calculate_trajectory, "moving-frame-orientation": movingframes}

    # Calculate remaining trajectory
    new_invars, calculate_trajectory, movingframes, tot_time_rot = FS_online_generation_problem2.generate_trajectory_OLD(model_invariants,boundary_constraints,new_stepsize,weight_params,initial_values)
    if use_fatrop_solver:
        print('')
        print("TOTAL time to generate new trajectory: ")
        print(str(tot_time_rot) + '[s]')

    # fig_traj.clf()
    # ax_traj = fig_traj.add_subplot(111, projection='3d')
    # ax_traj.view_init(elev=26, azim=-40)
    # ax_traj.plot(trajectory[:, 0], trajectory[:, 1], trajectory[:, 2], 'b')
    # ax_traj.plot(calculate_trajectory[:, 0], calculate_trajectory[:, 1], calculate_trajectory[:, 2], 'r')
    # fig_traj.canvas.draw()
    # fig_traj.canvas.flush_events()


    for i in range(len(calculate_trajectory)):
        calculate_trajectory[i] = orthonormalize(calculate_trajectory[i])

    # NEED TO WORK ON THE VISUALIZATION!!!####################################
    
    # # Visualization
    # clear_output(wait=True)
    

    # ax.plot(trajectory_position[:,0],trajectory_position[:,1],trajectory_position[:,2],'b')
    # for i in indx:
    #     plot_stl(opener_location,trajectory_position[i,:],calculate_trajectory[i,:,:],colour="c",alpha=0.2,ax=ax)
    # cut_trajectory_position = trajectory_position_iter[current_index:,:]
    # interp = ip.interp1d(np.linspace(0,1,len(cut_trajectory_position)),cut_trajectory_position, axis = 0)
    # trajectory_position_iter = interp(np.linspace(0,1,len(calculate_trajectory)))
    # indx_iter = np.trunc(np.linspace(0,len(calculate_trajectory)-1,n_frames))
    # indx_iter = indx_iter.astype(int)
    # for i in indx_iter:
    #     plot_3d_frame(trajectory_position_iter[i,:],calculate_trajectory[i,:,:],1,0.05,['red','green','blue'],ax)
    #     plot_stl(opener_location,trajectory_position_iter[i,:],calculate_trajectory[i,:,:],colour="r",alpha=0.2,ax=ax)
    
    # pl.plot_invariants(invariants,new_invars,arclength_n,progress_values,inv_type='FS_rot',fig=fig_invars)

    # if plt.get_backend() != 'agg':
    #     plt.show()
    
    old_progress = current_progress
    current_progress = old_progress + 1/window_len


# -*- coding: utf-8 -*-
"""
Created on Wed Mar 6 2024

note: urdf2casadi is not included in the invariants_py installation
install separately using pip install urdf2casadi or include it in the pyproject.toml file

@author: Riccardo
"""

import numpy as np
import casadi as cas
import rockit
import os
from math import pi
import invariants_py.data_handler as dh
import invariants_py.reparameterization as reparam
from forward_kinematics import forward_kinematics
from invariants_py.SE3 import rotate_z

def inv_kin(q_init, q_joint_lim, des_p_obj, des_R_obj, path_to_urdf, root = 'base_link', tip = 'tool0', window_len = 1, nb_joints = 6, fatrop_solver = False):

    ocp = rockit.Ocp(T=1.0)

    # Define system states
    q = ocp.state(nb_joints) # robot joints

    # Define controls
    qdot = ocp.control(nb_joints)

    # Boundary values
    q_lim = ocp.parameter(nb_joints)

    # Parameters
    p_obj_m = ocp.parameter(3,grid='control+')
    R_obj_m_x = ocp.parameter(3,1,grid='control+')
    R_obj_m_y = ocp.parameter(3,1,grid='control+')
    R_obj_m_z = ocp.parameter(3,1,grid='control+')
    R_obj_m = cas.horzcat(R_obj_m_x,R_obj_m_y,R_obj_m_z)

    # Boundary constraints
    for i in range(nb_joints):
        ocp.subject_to(-q_lim[i] <= (q[i] <= q_lim[i]))

    # Define derivatives
    ocp.set_der(q,qdot)

    # Forward kinematics
    p_obj, R_obj = forward_kinematics(q,path_to_urdf,root,tip)
    # Specify the objective
    e_pos = cas.dot(p_obj - p_obj_m,p_obj - p_obj_m)
    e_rot = cas.dot(R_obj_m.T @ R_obj - np.eye(3),R_obj_m.T @ R_obj - np.eye(3))
    objective = ocp.sum(e_pos + e_rot + 0.001*cas.dot(qdot,qdot),include_last = True)
    ocp.add_objective(objective)
    
    if fatrop_solver:
        ocp.method(rockit.external_method('fatrop' , N=window_len-1))
        # pass
    else:
        ocp.method(rockit.MultipleShooting(N=window_len-1))
        ocp.solver('ipopt', {'expand':True,'ipopt.max_iter':100})

    ocp.set_initial(q,q_init.T)

    ocp.set_initial(qdot,0.001*np.ones((nb_joints,window_len-1)))

    # Set values of boundary constraints
    ocp.set_value(q_lim,q_joint_lim)

    #Set values of parameters
    ocp.set_value(p_obj_m,des_p_obj[:window_len].T)
    ocp.set_value(R_obj_m_x,des_R_obj[:window_len,:,0].T)
    ocp.set_value(R_obj_m_y,des_R_obj[:window_len,:,1].T)
    ocp.set_value(R_obj_m_z,des_R_obj[:window_len,:,2].T)

    # Solve the NLP
    sol = ocp.solve_limited()

    # joint_val = sol.value(q)
    _,joint_val = sol.sample(q,grid='control')

    return joint_val


if __name__ == "__main__":
    data_location = dh.find_data_path('beer_1.txt')
    trajectory,time = dh.read_pose_trajectory_from_txt(data_location)
    pose,time_profile,arclength,nb_samples,stepsize = reparam.reparameterize_trajectory_arclength(trajectory)
    N = 100
    root_link_name = "base_link"
    tip_link_name = "TCP_frame"
    path_to_urdf = dh.find_data_path('ur10.urdf')
    startpos = [0.3056, 0.0635, 0.441]
    des_p_obj = pose[:,:3,3]  + startpos #[0.6818214 , 0.23448511, 0.39779707] * np.ones((N,3)) #
    des_R_obj = pose[:,:3,:3]
    # des_R_obj = np.zeros((N,3,3))
    # for i in range(N):
    #     des_R_obj[i] = pose[-1,:3,:3] #np.eye(3)
    q_init = [-pi, -2.27, 2.27, -pi/2, -pi/2, pi/4] * np.ones((N,6))
    q_joint_lim = [2*pi, 2*pi, pi, 2*pi, 2*pi, 2*pi]
    q = inv_kin(q_init, q_joint_lim, des_p_obj, des_R_obj, path_to_urdf, root_link_name, tip_link_name, N)

    print(q)

# =======================================
# Debugging =======================================
# =======================================
# p_obj, R_obj = forward_kinematics(q,path_to_urdf,root_link_name,tip_link_name)

# for i in range(N):
#     p_obj, R_obj = forward_kinematics(q[i],path_to_urdf,root_link_name,tip_link_name)

#     #e_pos = cas.dot(p_obj - des_p_obj[i],p_obj - des_p_obj[i])
#     e_rot = cas.dot(des_R_obj[i].T @ R_obj - np.eye(3),des_R_obj[i].T @ R_obj - np.eye(3))
    
#     #print(des_R_obj[0])
#     #print(R_obj)
#     print(e_rot)


import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d.art3d import Poly3DCollection
from scipy.spatial.transform import Rotation as R
from invariants_py.external.GJK import geometry as GJK_geometry
from invariants_py.external.GJK import compute_distance as GJK_compute_distance

def collision_detection_bottle(p_obj_demo, R_demo, position_bottle, opener_geom, tilting_rotx, tilting_roty, tilting_rotz, mode,ax):
    # How many iterations to allow for collision detection.
    iterationsAllowed = 6

    # Define the geometry of the bottle by creating three circles: top, middle, bottom
    N = 100 # number of points in each circle
    z_top = np.zeros(N) # the top of bottle is at z = 0, we consider N points
    r_top = np.full_like(z_top, 0.0145) # radius at the top is 0.0145
    z_bottom = np.ones((2,N)) # z_bottom describes the middle and bottom of the bottle
    z_bottom[0] = np.ones((1,N))*(-0.087) # the middle of the bottle is at z = -0.087
    z_bottom[1] = np.ones((1,N))*(-0.174) # the bottom of the bottle is at z = -0.174
    r_bottom = np.full_like(z_top, 0.035) # radius from the middle to the bottom is 0.035

    theta = np.linspace(0, 2.*np.pi, N) # theta goes from 0 to 2*pi in N steps
    x_top = r_top * np.cos(theta) # x coordinate of all N circle points at the top
    y_top = r_top * np.sin(theta) # y coordinate of all N circle points at the top
    x_bottom = r_bottom * np.cos(theta) # x coordinate of all N circle points at the bottom/middle
    y_bottom = r_bottom * np.sin(theta) # y coordinate of all N circle points at the bottom/middle

    v_bottle = np.zeros((2*2*len(x_top),3)) # v_bottle is the matrix that contains all cylinder points, 2*2*len(x_top) is because I count middle cirlce twice + top and bottom
    k = 0
    for i in range(len(x_top)): # put data of cylinder in one matrix
        for j in range(2): # sequence is: top, middle,top,middle [...], middle,bottom,middle,bottom [...]
            if j == 0:
                v_bottle[k,:] = [x_top[i], y_top[i], z_top[i]]
            else:
                v_bottle[k,:] = [x_bottom[i], y_bottom[i], z_bottom[0,i]]
            v_bottle[k+2*len(x_top),:] = [x_bottom[i], y_bottom[i], z_bottom[j,i]]
            k += 1

    v_bottle += position_bottle # shift the bottle to the correct position

    F1 = np.zeros((2*(len(x_top)-1),4)) # F1 are contains sequence to create faces
    for i in range(len(F1)):
        F1[i,:] = [1+2*i, 2+2*i, 4+2*i, 3+2*i]

    # ax.add_collection3d(Poly3DCollection(v_bottle[F1.astype(int)])) # plot the shell of the bottle

    # if mode == 'rpy':
    #     ax.view_init(elev=tilting_rotx, azim=tilting_roty)
    # else:
    #     ax.view_init(elev=tilting_rotz, azim=np.rad2deg(np.arctan2(tilting_roty, tilting_rotz)))

    collision_sample = 0
    collision_happened = 0
    last_collision_sample = 0

    N_checks = 100 # number of checks for collision
    bo_dist = np.zeros((len(p_obj_demo),N_checks)) # matrix to store distances between bottle and opener
    for j in np.linspace(0, len(p_obj_demo)-1, N_checks).astype(int):
        fm = [0, 1, 2, 3]
        v_opener = np.tile(p_obj_demo[j,:], (4,1)) + np.dot(R_demo[j,:,:], opener_geom[[0,18,20,2],:].T).T # include opener geometry
        ax.add_collection3d(Poly3DCollection([v_opener[fm]], color='r',alpha=0.1))

        b = GJK_geometry.Polytope(v_bottle)
        o = GJK_geometry.Polytope(v_opener)

        set_GJK = GJK_compute_distance.ComputeDist([("bottle",b), ("opener",o)])
        bo_dist[j,:] = set_GJK.GetDist("bottle", "opener")
        if np.any(bo_dist[j,:] < 0.00001):
            if collision_sample == 0:
                collision_sample = j
            else:
                last_collision_sample = j
            collision_happened = 1
    
    return collision_happened, collision_sample, last_collision_sample


"""
Check all examples, tests, modules and notebooks to see if they run successfully
"""

import os
import subprocess

# Directories containing the examples and tests
directories = ["examples", "tests", "invariants_py", "notebooks"]
#directories = ["notebooks"]

# Initialize an empty list to hold examples and tests that fail to execute
failed_scripts = []

# Set the matplotlib backend to Agg in the environment variable to prevent showing plots
my_env = os.environ.copy()
my_env["MPLBACKEND"] = "agg"

# Get the directory containing this script
script_dir = os.path.dirname(os.path.realpath(__file__))

# Function to run scripts in a directory
def run_scripts(directory):
    failed_scripts_local = []
    directory_path = os.path.join(script_dir, directory)
    for root, dirs, files in os.walk(directory_path):
        # Get a list of all Python scripts in the current directory
        scripts = [f for f in files if f.endswith('.py') or f.endswith('.ipynb')]

        # Iterate over each script
        for script in scripts:
            # Construct the full script path
            script_path = os.path.join(root, script)

            # Print the name of the script being run
            print(f"Running script: {script_path}")

            # Try to execute the script
            try:
                if script.endswith('.py'):
                    subprocess.check_output(["python", script_path], env=my_env)
                elif script.endswith('.ipynb'):
                    subprocess.check_output(["jupyter", "nbconvert", "--execute", "--to", "notebook", script_path], env=my_env)
                    # Delete the generated notebook
                    os.remove(os.path.splitext(script_path)[0] + '.nbconvert.ipynb')
            except subprocess.CalledProcessError:
                # If an error occurs, add the script to the list of failed scripts
                failed_scripts_local.append(script_path)

    return failed_scripts_local

# Run scripts for each directory
for directory in directories:
    failed_scripts.extend(run_scripts(directory))

# Print out the scripts that failed to execute
print("")
print("========================================")
if failed_scripts:
    print("The following scripts failed to execute:")
    for script in failed_scripts:
        print(script)
    print("========================================")
    print("")
    raise Exception('One or more scripts failed, check the above output to see which ones.')
else:
    print("All scripts executed successfully.")
    print("========================================")
    print("")
    


# -*- coding: utf-8 -*-
"""
Created on Tue Mar 19 2024

@author: Riccardo
"""
import os
import urdf2casadi.urdfparser as u2c
from invariants_py.SE3 import rotate_z
from math import pi

def forward_kinematics(q, path_to_urdf, root = 'base_link', tip = 'tool0'):
    ur10 = u2c.URDFparser()
    ur10.from_file(path_to_urdf)
    fk_dict = ur10.get_forward_kinematics(root, tip)
    forward_kinematics = fk_dict["T_fk"]
    T_rob = forward_kinematics(q.T)
    T_rob = rotate_z(pi) @ T_rob
    p_obj = T_rob[:3,3]
    R_obj = T_rob[:3,:3]

    return p_obj, R_obj


# Generate translation trajectory from model invariants

# Import necessary modules
from invariants_py import data_handler as dh
import invariants_py.plotters as plotters
from invariants_py.opti_generate_position_from_vector_invariants import generate_trajectory_translation

# Find the path to the data file
path_to_data = dh.find_data_path("sine_wave_invariant_model.csv")

# Load the invariants data from the file
invariant_model = dh.read_invariants_from_csv(path_to_data)

# Specify the boundary constraints
boundary_constraints = {"position": {"initial": [0, 0, 0], "final": [2.5, 0.25, 0.5]}}

# Calculate the translation trajectory given the invariants data
invariants, trajectory, mf, progress_values = generate_trajectory_translation(invariant_model, boundary_constraints)

# Plot the boundary constraints and the calculated trajectory
plotters.plot_trajectory_and_bounds(boundary_constraints, trajectory)

# Plot the invariant model and the calculated invariants
plotters.compare_invariants(invariants, invariant_model[:,1:], progress_values, invariant_model[:,0])


# Calculate the invariants of a translation trajectory

# Import necessary modules
from invariants_py import data_handler as dh
import invariants_py.plotters as plotters
from invariants_py import calculate_invariants

# Find the path to the data file
path_data = dh.find_data_path("sine_wave.txt") # TODO convert sine_wave.txt to sine_wave.csv for consistency?

# Load the trajectory data from the file
# todo format of data
trajectory, time = dh.read_pose_trajectory_from_txt(path_data)

# Calculate the invariants of the translation trajectory
invariants, progress, calc_trajectory = calculate_invariants.calculate_invariants_translation(trajectory)

# Plot the calculated invariants and corresponding trajectory
plotters.plot_invariants_new(invariants, progress)
plotters.plot_trajectory(calc_trajectory)

# Save invariant model to a file
dh.save_invariants_to_csv(progress, invariants, "sine_wave_invariants.csv")

#input("Press Enter to continue...")


"""
Implements the differential equations and discrete dynamics of the vector and screw invariants
"""

import casadi as cas
import numpy as np
#import invariants_py.SE3 as SE3

def rodriguez_rot_form(omega,h):
    """Return a rotation matrix which is the result of rotating with omega over time interval h"""
    omega_norm = cas.norm_2(omega)
    return np.eye(3) + cas.sin(omega_norm*h)/omega_norm*cas.skew(omega) + (1-cas.cos(omega_norm*h))/omega_norm**2 * cas.mtimes(cas.skew(omega),cas.skew(omega))
    #return cas.SX.eye(3) + cas.sin(omega_norm*h)/omega_norm*cas.skew(omega) + (1-cas.cos(omega_norm*h))/omega_norm**2 * cas.mtimes(cas.skew(omega),cas.skew(omega))

#def rodriguez_rot_form3(deltatheta):
#    theta = deltatheta[0,1]
#    return cas.SX.eye(3) + cas.mtimes(cas.sin(theta),K) + (1-cas.cos(theta))*cas.mtimes(K,K)

def rodrigues2(skewer):
    
    omega = cas.vertcat(-skewer[1,2],skewer[0,2],-skewer[0,1])
    return rodriguez_rot_form(omega,h=1)

def geo_integrator(R_t, R_r, R_obj, p_obj, u, h):
    """Integrate invariants over interval h starting from a current state (object pose + moving frames)"""
    # Define a geometric integrator for eFSI,
    # (meaning rigid-body motion is perfectly integrated assuming constant invariants)
    i1 = u[0]
    i2 = u[1]
    i3 = u[2]
    i4 = u[3]
    i5 = u[4]
    i6 = u[5]

    omega_t = cas.vertcat(i6,i5,0)
    omega_r = cas.vertcat(i3,i2,0)
    omega_o = cas.mtimes(R_r, cas.vertcat(i1,0,0))

    #translation
    deltaR_t = rodriguez_rot_form(omega_t,h)
    R_t_plus1 = cas.mtimes(R_t,deltaR_t)
    p_obj_plus1 = cas.mtimes(R_t, cas.vertcat(i4,0,0))*h + p_obj

    #rotation
    deltaR_r = rodriguez_rot_form(omega_r,h)
    R_r_plus1 = cas.mtimes(R_r,deltaR_r)

    deltaR_o = rodriguez_rot_form(omega_o,h)
    R_obj_plus1 = cas.mtimes(deltaR_o,R_obj)

    return (R_t_plus1, R_r_plus1, R_obj_plus1, p_obj_plus1)



def vector_invariants_position(R_t, p_obj, u, h):
    """
    Discrete dynamics of the vector invariants for position. Integrate invariants over interval h starting from a current state (object pose + moving frames)
    """
    # Define a geometric integrator for eFSI,
    # (meaning rigid-body motion is perfectly integrated assuming constant invariants)
    
    
    # object translation speed
    # curvature speed translational Frenet-Serret
    # torsion speed translational Frenet-Serret
    
    i4 = u[0]
    i5 = u[1]
    i6 = u[2]

    omega = cas.vertcat(i6,0,i5)
    omega_norm = cas.norm_2(omega)
    v = cas.vertcat(i4,0,0)

    deltaR = np.eye(3) + cas.sin(omega_norm @ h)/omega_norm*cas.skew(omega) + (1-cas.cos(omega_norm @ h))/omega_norm**2 * cas.mtimes(cas.skew(omega),cas.skew(omega))
    deltaP = (np.eye(3)-deltaR) @ cas.skew(omega) @ v/omega_norm**2 + omega @ omega.T @ v/omega_norm**2*h
    
    R_t_plus1 = R_t @ deltaR
    p_obj_plus1 = R_t @ deltaP + p_obj

    return (R_t_plus1, p_obj_plus1)

def dyn_vector_invariants_rotation(R_r, R_obj, u, h):
    """Integrate invariants over interval h starting from a current state (object pose + moving frames)"""
    # Define a geometric integrator for eFSI,
    # (meaning rigid-body motion is perfectly integrated assuming constant invariants)
    
    
    # object rotational speed
    # curvature speed rotational Frenet-Serret
    # torsion speed rotational Frenet-Serret
    
    i1 = u[0]
    i2 = u[1]
    i3 = u[2]

    omega = cas.vertcat(i3,0,i2) # THIS FOLLOWS THE FORMULATION OF NEW OCP (look at integrator_vector_invariants_to_rot.m)
    v = cas.vertcat(i1,0,0)
    v_world = R_r @ v

    omega_norm = cas.norm_2(omega)
    omega_norm = cas.if_else(omega_norm == 0.0, 0.00001, omega_norm)
    e_omega = omega/omega_norm
    deltaR_r = np.eye(3) + cas.sin(omega_norm @ h)*cas.skew(e_omega) + (1-cas.cos(omega_norm @ h)) * cas.mtimes(cas.skew(e_omega),cas.skew(e_omega))
    v_norm = cas.norm_2(v_world) 
    v_norm = cas.if_else(v_norm == 0.0, 0.00001, v_norm)
    e_v = v_world/v_norm
    deltaR_obj = np.eye(3) + cas.sin(v_norm @ h)*cas.skew(e_v) + (1-cas.cos(v_norm @ h)) * cas.mtimes(cas.skew(e_v),cas.skew(e_v))

    R_r_plus1 = R_r @ deltaR_r
    R_obj_plus1 = deltaR_obj @ R_obj

    return (R_r_plus1, R_obj_plus1)

def dyn_vector_invariants_rotation_sequential(R_r, R_obj, u, h):
    """Integrate invariants over interval h starting from a current state (object pose + moving frames)"""
    # Define a geometric integrator for eFSI,
    # (meaning rigid-body motion is perfectly integrated assuming constant invariants)
    
    
    # object rotational speed
    # curvature speed rotational Frenet-Serret
    # torsion speed rotational Frenet-Serret
    
    # TO DO: make this implementation more compact
    i1 = u[0]*h
    i2 = u[1]*h
    i3 = u[2]*h
    
    rot_x_obj = cas.MX.zeros(3,3)
    rot_x_obj[0,0] = 1
    rot_x_obj[1,1] = cas.cos(i1)
    rot_x_obj[2,2] = cas.cos(i1)
    rot_x_obj[1,2] = -cas.sin(i1)
    rot_x_obj[2,1] = cas.sin(i1)
    
    rot_z = cas.MX.zeros(3,3)
    rot_z[0,0] = cas.cos(i2)
    rot_z[1,1] = cas.cos(i2)
    rot_z[0,1] = -cas.sin(i2)
    rot_z[1,0] = cas.sin(i2)
    rot_z[2,2] = 1
    
    rot_x = cas.MX.zeros(3,3)
    rot_x[0,0] = 1
    rot_x[1,1] = cas.cos(i3)
    rot_x[2,2] = cas.cos(i3)
    rot_x[1,2] = -cas.sin(i3)
    rot_x[2,1] = cas.sin(i3)

    deltaR = rot_z @ rot_x
    
    R_obj_plus1 = (R_r @ rot_x_obj @ R_r.T) @ R_obj
    R_r_plus1 = R_r @ deltaR

    return (R_r_plus1, R_obj_plus1)

def rodriguez_rot_form2(omega,h):
    """Return a rotation matrix which is the result of rotating with omega over time interval h"""
    omega_norm = cas.norm_2(omega)
    #return np.eye(3) + sin(omega_norm*h)/omega_norm*skew(omega) + (1-cos(omega_norm*h))/omega_norm**2 * mtimes(skew(omega),skew(omega))
    return np.eye(3) + cas.sin(omega_norm*h)/omega_norm*cas.skew(omega) + (1-cas.cos(omega_norm*h))/omega_norm**2 * cas.mtimes(cas.skew(omega),cas.skew(omega))

def geo_integrator_eFSI(R_t, R_r, R_obj, p_obj, u, h):
    """Integrate invariants over interval h starting from a current state (object pose + moving frames)"""
    #SAME ONE BUT THIS ONE USES ACTUAL VALUES INSTEAD OF CASADI VARIABLES
    # Define a geometric integrator for eFSI,
    # (meaning rigid-body motion is perfectly integrated assuming constant invariants)
    i1 = u[0]
    i2 = u[1]
    i3 = u[2]
    i4 = u[3]
    i5 = u[4]
    i6 = u[5]

    omega_t = np.array([[i6],[i5],[0]])
    omega_r = np.array([[i3],[i2],[0]])
    omega_o = cas.mtimes(R_r, np.array([[i1],[0],[0]]))

    #translation
    deltaR_t = rodriguez_rot_form2(omega_t,h)
    R_t_plus1 = cas.mtimes(R_t,deltaR_t)
    p_obj_plus1 = cas.mtimes(R_t, np.array([[i4],[0],[0]]))*h + p_obj

    #rotation
    deltaR_r = rodriguez_rot_form2(omega_r,h)
    R_r_plus1 = cas.mtimes(R_r,deltaR_r)

    deltaR_o = rodriguez_rot_form2(omega_o,h)
    R_obj_plus1 = cas.mtimes(deltaR_o,R_obj)

    return (R_t_plus1, R_r_plus1, R_obj_plus1, p_obj_plus1)

def offset_integrator(invariants, h):
    """Return offset in object position and rotation by reconstructing the invariant signature"""
    R_t_0 = np.eye(3)
    R_r_0 = np.eye(3)
    R_offset = np.eye(3)
    p_offset = np.array([[0],[0],[0]])
    for i in range(np.shape((invariants))[1]):
        (R_t_0, R_r_0, R_offset, p_offset) \
            = geo_integrator_eFSI(R_t_0, R_r_0, R_offset, p_offset, invariants[:,i], h)
    return (R_offset, p_offset)


def define_geom_integrator_tra_FSI_casadi(h):
    ## Generate optimal eFSI trajectory
    # System states
    R_t  = cas.MX.sym('R_t',3,3) # translational Frenet-Serret frame
    p_obj = cas.MX.sym('p_obj',3,1) # object position
    x = cas.vertcat(cas.vec(R_t), p_obj)
    #np = length(R_obj(:)) + length(p_obj)

    # System controls (invariants)
    u = cas.MX.sym('i',3,1)

    ## Define a geometric integrator for eFSI, (meaning rigid-body motion is perfectly integrated assuming constant invariants)
    (R_t_plus1, p_obj_plus1) = vector_invariants_position(R_t, p_obj, u, h)
    out_plus1 = cas.vertcat(cas.vec(R_t_plus1),  p_obj_plus1)
    integrator = cas.Function("phi", [x,u,h] , [out_plus1])
    
    return integrator

def define_geom_integrator_rot_FSI_casadi(h):
    ## Generate optimal eFSI trajectory
    # System states
    R_r  = cas.MX.sym('R_r',3,3) # rotational Frenet-Serret frame
    R_obj = cas.MX.sym('R_obj',3,3) # object orientation
    x = cas.vertcat(cas.vec(R_r), cas.vec(R_obj))

    # System controls (invariants)
    u = cas.MX.sym('i',3,1)

    ## Define a geometric integrator for eFSI, (meaning rigid-body motion is perfectly integrated assuming constant invariants)
    (R_r_plus1, R_obj_plus1) = dyn_vector_invariants_rotation(R_r, R_obj, u, h)
    out_plus1 = cas.vertcat(cas.vec(R_r_plus1),  cas.vec(R_obj_plus1))
    integrator = cas.Function("phi", [x,u,h] , [out_plus1])
    
    return integrator



# Imports
import numpy as np
import invariants_py.data_handler as dh
import matplotlib.pyplot as plt
import invariants_py.reparameterization as reparam
from invariants_py.opti_calculate_vector_invariants_position_mf import OCP_calc_pos
from invariants_py.opti_generate_position_from_vector_invariants import OCP_gen_pos
import invariants_py.spline_handler as spline_handler
import invariants_py.plotters as plotters

"""Input data"""

data_location = dh.find_data_path("sine_wave.txt")
trajectory,time = dh.read_pose_trajectory_from_txt(data_location)
pose,time_profile,arclength,nb_samples,stepsize = reparam.reparameterize_trajectory_arclength(trajectory)
arclength_n = arclength/arclength[-1]
trajectory = pose[:,0:3,3]
plotters.plot_trajectory_test(trajectory)

"""Calculate invariants"""

# Symbolic specification
FS_calculation_problem = OCP_calc_pos(window_len=nb_samples, bool_unsigned_invariants = False, w_pos = 1, w_deriv = (10**-5)*np.array([1.0, 1.0, 1.0]), w_abs = (10**-6)*np.array([1.0, 1.0]))

# Calculate invariants given measurements
invariants, calculate_trajectory, movingframes = FS_calculation_problem.calculate_invariants_global(trajectory,stepsize)

init_vals_calculate_trajectory = calculate_trajectory
init_vals_movingframes = movingframes

fig = plt.figure(figsize=(8,8))
ax = fig.add_subplot(111, projection='3d')
ax = plt.axes(projection='3d')
ax.plot(trajectory[:,0],trajectory[:,1],trajectory[:,2],'.-')
ax.plot(calculate_trajectory[:,0],calculate_trajectory[:,1],calculate_trajectory[:,2],'.-')

f, (ax1, ax2, ax3) = plt.subplots(1, 3, sharey=True, figsize=(10,3))
ax1.plot(arclength_n,invariants[:,0])
ax1.set_title('Velocity [m/m]')
ax2.plot(arclength_n,invariants[:,1])
ax2.set_title('Curvature [rad/m]')
ax3.plot(arclength_n,invariants[:,2])
ax3.set_title('Torsion [rad/m]')

# Spline of model invariants
spline_model_invariants = spline_handler.create_spline_model(arclength_n, invariants)

#%% 
current_progress = 0.4
number_samples = 40
calculate_trajectory = init_vals_calculate_trajectory
movingframes = init_vals_movingframes

progress_values = np.linspace(current_progress, arclength_n[-1], number_samples)
model_invariants,new_stepsize = spline_handler.interpolate_invariants(spline_model_invariants,progress_values)

f, (ax1, ax2, ax3) = plt.subplots(1, 3, sharey=True, figsize=(10,3))
ax1.plot(arclength_n,invariants[:,0])
ax1.plot(progress_values,model_invariants[:,0],'r.')
ax1.set_title('Velocity [m/m]')
ax2.plot(arclength_n,invariants[:,1])
ax2.plot(progress_values,model_invariants[:,1],'r.')
ax2.set_title('Curvature [rad/m]')
ax3.plot(arclength_n,invariants[:,2])
ax3.plot(progress_values,model_invariants[:,2],'r.')
ax3.set_title('Torsion [rad/m]')


# new constraints
current_index = round(current_progress*len(trajectory))
p_obj_start = calculate_trajectory[current_index]
p_obj_end = calculate_trajectory[-1] - np.array([-0.2, 0.0, 0.0])
R_FS_start = movingframes[current_index]
R_FS_end = movingframes[-1]


# specify optimization problem symbolically
FS_online_generation_problem = OCP_gen_pos(N=number_samples,w_invars = 10**2*np.array([10**1, 1.0, 1.0]))

# Solve
new_invars, new_trajectory, new_movingframes = FS_online_generation_problem.generate_trajectory(U_demo = model_invariants, p_obj_init = calculate_trajectory, R_t_init = movingframes, R_t_start = R_FS_start, R_t_end = R_FS_end, p_obj_start = p_obj_start, p_obj_end = p_obj_end, step_size = new_stepsize)


fig = plt.figure(figsize=(14,8))
ax = fig.add_subplot(111, projection='3d')
ax.plot(trajectory[:,0],trajectory[:,1],trajectory[:,2],'b')
ax.plot(new_trajectory[:,0],new_trajectory[:,1],new_trajectory[:,2],'r')

fig = plt.figure()
plt.subplot(1,3,1)
plt.plot(progress_values,new_invars[:,0],'r')
plt.plot(arclength_n,invariants[:,0],'b')
plt.plot(0,0)
plt.title('Velocity [m/m]')

plt.subplot(1,3,2)
plt.plot(progress_values,(new_invars[:,1]),'r')
plt.plot(arclength_n,invariants[:,1],'b')
plt.plot(0,0)
plt.title('Curvature [rad/m]')

plt.subplot(1,3,3)
plt.plot(progress_values,(new_invars[:,2]),'r')
plt.plot(arclength_n,invariants[:,2],'b')
plt.plot(0,0)
plt.title('Torsion [rad/m]')

if plt.get_backend() != 'agg':
    plt.show()


#%% Visualization

window_len = 20

# specify optimization problem symbolically
FS_online_generation_problem = OCP_gen_pos(N=window_len,w_invars = 10**1*np.array([10**1, 1.0, 1.0]))



current_progress = 0.0
old_progress = 0.0

calculate_trajectory = init_vals_calculate_trajectory
movingframes = init_vals_movingframes

plt.ion() # interactive plotting, necessary for updating the plot
fig = plt.figure(figsize=(14, 8)) # plot showing update of progress

while current_progress <= 1.0:
    
    print(f"current progress = {current_progress}")

    # Resample invariants for current progress
    progress_values = np.linspace(current_progress, arclength_n[-1], window_len)
    model_invariants,new_stepsize = spline_handler.interpolate_invariants(spline_model_invariants,progress_values)
    
    # Boundary constraints
    current_index = round( (current_progress - old_progress) * len(calculate_trajectory))
    #print(current_index)
    p_obj_start = calculate_trajectory[current_index]
    p_obj_end = trajectory[-1] - current_progress*np.array([-0.2, 0.0, 0.0])
    R_FS_start = movingframes[current_index] 
    R_FS_end = movingframes[-1] 

    # Calculate remaining trajectory
    new_invars, calculate_trajectory, movingframes = FS_online_generation_problem.generate_trajectory(U_demo = model_invariants, p_obj_init = calculate_trajectory, R_t_init = movingframes, R_t_start = R_FS_start, R_t_end = R_FS_end, p_obj_start = p_obj_start, p_obj_end = p_obj_end, step_size = new_stepsize)

    # Visualization
    #clear_output(wait=True)
    
    #fig = plt.figure(figsize=(14,8))
    #ax = fig.add_subplot(231, projection='3d')
    #ax.view_init(elev=26, azim=140)
    #ax.plot(trajectory[:,0],trajectory[:,1],trajectory[:,2],'b')
    #ax.plot(calculate_trajectory[:,0],calculate_trajectory[:,1],#calculate_trajectory[:,2],'r')
    
    plotters.plot_trajectory_invariants_online(arclength_n, invariants, progress_values, new_invars, fig)
    
    old_progress = current_progress
    current_progress = old_progress + 1/window_len
    
    



from invariants_py import data_handler as dh
import invariants_py.invariant_descriptor_class as invars
#import numpy as np

# Initialization
data_location = dh.find_data_path("sine_wave.txt")
parameterization = 'timebased' # {timebased,geometric}

"""
Part 1: calculation invariants
"""
# Calculate invariants + return corresponding trajectory, sample period is found from timestamps
descriptor = invars.MotionTrajectory(data_location, invariantType=parameterization, suppressPlotting=True)
poses = descriptor.getPosesFromInvariantSignature(descriptor.getInvariantSignature())
invariants = descriptor.getInvariantsDemo()['U']


"""
Part 2: generation new trajectories
"""
newposes, newinvariants = descriptor.generateMotionFromInvariants(startPose=poses[0],endPose=poses[-1])


from pathlib import Path

import numpy as np
import matplotlib.pyplot as plt
import invariants_py.reparameterization as reparam
import invariants_py.opti_calculate_vector_invariants_position_mf as FS1
import invariants_py.opti_calculate_vector_invariants_position as FS2
#import invariants_py.opti_calculate_vector_invariants_position_mj as FS3
from invariants_py import data_handler as dh
#%%

data_location = dh.find_data_path('contour_coordinates.out')


position_data = np.loadtxt(data_location, dtype='float')
trajectory,time_profile,arclength,nb_samples,stepsize = reparam.reparameterize_positiontrajectory_arclength(position_data)
stepsize_orig = stepsize
arclength_n = arclength/arclength[-1]

plt.figure(figsize=(8,3))
plt.axis('equal')
plt.plot(trajectory[:,0],trajectory[:,1],'.-')



"""
Old optimization problem
"""

#%%
# specify optimization problem symbolically
FS_calculation_problem = FS1.OCP_calc_pos(window_len=nb_samples, bool_unsigned_invariants = True, w_pos = 100, w_deriv = (10**-7)*np.array([1.0, 1.0, 1.0]), w_abs = (10**-5)*np.array([1.0, 1.0]))

# calculate invariants given measurements
invariants, calculate_trajectory, movingframes = FS_calculation_problem.calculate_invariants_global(trajectory,stepsize)

# figures
plt.figure(figsize=(14,6))
plt.subplot(2,2,1)
plt.plot(trajectory[:,0],trajectory[:,1],'.-')
plt.plot(calculate_trajectory[:,0],calculate_trajectory[:,1],'.-')
plt.title('Trajectory')

plt.subplot(2,2,3)
plt.plot(arclength_n,invariants[:,0])
plt.plot(0,0)
plt.title('Velocity [m/-]')

plt.subplot(2,2,2)
plt.plot(arclength_n,invariants[:,1])
plt.plot(0,0)
plt.title('Curvature [rad/-]')

plt.subplot(2,2,4)
plt.plot(arclength_n,invariants[:,2])
plt.plot(0,1)
plt.title('Torsion [rad/-]')

if plt.get_backend() != 'agg':
    plt.show()



# """ 
# OCP 2: Minimum-jerk optimization problem
# """

# #%%
# # specify optimization problem symbolically
# FS_calculation_problem = FS3.OCP_calc_pos(window_len=nb_samples, w_pos = 100, w_regul = 10**-9)

# # calculate invariants given measurements
# invariants, calculate_trajectory, movingframes = FS_calculation_problem.calculate_invariants_global(trajectory,stepsize)

# # figures
# plt.figure(figsize=(14,6))
# plt.subplot(2,2,1)
# plt.plot(trajectory[:,0],trajectory[:,1],'.-')
# plt.plot(calculate_trajectory[:,0],calculate_trajectory[:,1],'.-')

# plt.subplot(2,2,3)
# plt.plot(arclength_n,invariants[:,0])
# plt.plot(0,0)
# plt.title('Velocity [m/-]')

# plt.subplot(2,2,2)
# plt.plot(arclength_n,invariants[:,1])
# plt.plot(0,0)
# plt.title('Curvature [rad/-]')

# plt.subplot(2,2,4)
# plt.plot(arclength_n,invariants[:,2])
# plt.plot(0,1)
# plt.title('Torsion [rad/-]')

# if plt.get_backend() != 'agg':
    plt.show()


""" 
Reformulated optimization problem
"""

#%%
# specify optimization problem symbolically
FS_calculation_problem = FS2.OCP_calc_pos(window_len=nb_samples, bool_unsigned_invariants = False, rms_error_traj = 0.001)

# calculate invariants given measurements
invariants, calculate_trajectory, movingframes = FS_calculation_problem.calculate_invariants_global(trajectory,stepsize)

# figures
plt.figure(figsize=(14,6))
plt.subplot(2,2,1)
plt.plot(trajectory[:,0],trajectory[:,1],'.-')
plt.plot(calculate_trajectory[:,0],calculate_trajectory[:,1],'.-')
plt.title('Trajectory')

plt.subplot(2,2,3)
plt.plot(arclength_n,invariants[:,0])
plt.plot(0,0)
plt.title('Velocity [m/-]')

plt.subplot(2,2,2)
plt.plot(arclength_n,invariants[:,1])
plt.plot(0,0)
plt.title('Curvature [rad/-]')

plt.subplot(2,2,4)
plt.plot(arclength_n,invariants[:,2])
plt.plot(0,1)
plt.title('Torsion [rad/-]')

if plt.get_backend() != 'agg':
    plt.show()


from invariants_py.opti_generate_position_from_vector_invariants import OCP_gen_pos
import numpy as np
import invariants_py.spline_handler as sh
from invariants_py.reparameterization import interpR
from invariants_py.initialization import FSr_init

def generate_initvals_from_bounds(boundary_constraints,N):
    
    # Generate initial trajectory using linear interpolation
    p1 = boundary_constraints["position"]["final"]
    p0 = boundary_constraints["position"]["initial"]
    initial_trajectory = np.linspace(p0, p1, N)

    # Generate corresponding initial invariants
    diff_vector = np.array(p1) - np.array(p0)
    L = np.linalg.norm(diff_vector)
    initial_invariants = np.tile(np.array([L,0,0]),(N,1))

    # Generate corresponding initial moving frames using Gram-Schmidt process
    e_x = diff_vector / L
    e_y = np.array([0, 1, 0]) - np.dot(np.array([0, 1, 0]), e_x) * e_x
    e_y = e_y / np.linalg.norm(e_y)
    e_z = np.cross(e_x, e_y)
    R_mf = np.column_stack((e_x, e_y, e_z))
    initial_movingframes = np.tile(R_mf, (N,1,1))

    initial_values = {
        "trajectory": initial_trajectory,
        "moving-frames": initial_movingframes,
        "invariants": initial_invariants
    }

    return initial_values

def generate_trajectory_translation(invariant_model, boundary_constraints, N=40):
    
    # Specify optimization problem symbolically
    OCP = OCP_gen_pos(N = N)

    # Initial values
    initial_values = generate_initvals_from_bounds(boundary_constraints, N)

    # Resample model invariants to desired number of N samples
    spline_invariant_model = sh.create_spline_model(invariant_model[:,0], invariant_model[:,1:])
    progress_values = np.linspace(invariant_model[0,0],invariant_model[-1,0],N)
    model_invariants,progress_step = sh.interpolate_invariants(spline_invariant_model, progress_values)
    
    # Calculate remaining trajectory
    invariants, trajectory, mf = OCP.generate_trajectory_global(model_invariants,initial_values,boundary_constraints,progress_step)

    return invariants, trajectory, mf, progress_values

        
        



import numpy as np
from math import pi
import casadi as cas
import rockit
from invariants_py.dynamics_invariants import dyn_vector_invariants_rotation
from invariants_py import SO3
from invariants_py.initialization import initialize_VI_rot
from invariants_py.ocp_helper import check_solver, tril_vec

class OCP_calc_rot:

    def __init__(self, window_len = 100, bool_unsigned_invariants = False, rms_error_traj = 4*pi/180, fatrop_solver = False):
       
        ocp = rockit.Ocp(T=1.0) # create optimization problem
        N = window_len # number of samples in the window

        fatrop_solver = check_solver(fatrop_solver)       
        #%% Create decision variables and parameters for the optimization problem
        
        # Define system states X (unknown object orientation + moving frame orientation at every time step) 
        R_obj_x = ocp.state(3,1) # object orientation, first axis
        R_obj_y = ocp.state(3,1) # object orientation, second axis
        R_obj_z = ocp.state(3,1) # object orientation, third axis
        R_obj = cas.horzcat(R_obj_x,R_obj_y,R_obj_z)
        R_r_x  = ocp.state(3,1) # rotational Frenet-Serret frame, first axis
        R_r_y  = ocp.state(3,1) # rotational Frenet-Serret frame, second axis
        R_r_z  = ocp.state(3,1) # rotational Frenet-Serret frame, third axis
        R_r = cas.horzcat(R_r_x,R_r_y,R_r_z)
        
        # Define system controls (invariants at every time step)
        invars = ocp.control(3) # three invariants [angular velocity | curvature rate | torsion rate]

        # Define system parameters P (known values in optimization that need to be set right before solving)
        R_obj_m_x = ocp.parameter(3,1,grid='control+') # measured object orientation, first axis
        R_obj_m_y = ocp.parameter(3,1,grid='control+') # measured object orientation, second axis
        R_obj_m_z = ocp.parameter(3,1,grid='control+') # measured object orientation, third axis
        R_obj_m = cas.horzcat(R_obj_m_x,R_obj_m_y,R_obj_m_z) 
        # ocp.parameter(3,3,grid='control',include_last=True)#
        # R_r_0 = ocp.parameter(3,3) # THIS IS COMMENTED OUT IN MATLAB, WHY?
        
        h = ocp.parameter(1) # stepsize
        
        # Define system discrete dynamics (integrate current state + controls to obtain next state)
        (R_r_plus1, R_obj_plus1) = dyn_vector_invariants_rotation(R_r, R_obj, invars, h)
        ocp.set_next(R_obj,R_obj_plus1)
        ocp.set_next(R_r,R_r_plus1)
            
        #%% Specifying the constraints
        
        # Constrain rotation matrices to be orthogonal (only needed for one timestep, property is propagated by integrator)
        ocp.subject_to(ocp.at_t0(tril_vec(R_obj.T @ R_obj - np.eye(3))==0.))
        ocp.subject_to(ocp.at_t0(tril_vec(R_r.T @ R_r - np.eye(3))==0.))

        # Lower bounds on controls
        if bool_unsigned_invariants:
            ocp.subject_to(invars[0,:]>=0) # lower bounds on control
            ocp.subject_to(invars[1,:]>=0) # lower bounds on control

        # Measurement fitting constraint
        ek = cas.dot(R_obj_m.T @ R_obj - np.eye(3),R_obj_m.T @ R_obj - np.eye(3)) # squared rotation error
        
        running_ek = ocp.state() # running sum of squared error
        ocp.subject_to(ocp.at_t0(running_ek == 0))
        ocp.set_next(running_ek, running_ek + ek)

        total_ek = ocp.state() # total sum of squared error
        ocp.set_next(total_ek, total_ek)
        ocp.subject_to(ocp.at_tf(total_ek == running_ek + ek))
        
        total_ek_scaled = total_ek/N/rms_error_traj**2 # scaled total error
        ocp.subject_to(total_ek_scaled < 1)

        # opti.subject_to(U[1,-1] == U[1,-2]); # Last sample has no impact on RMS error ##### HOW TO ACCESS U[1,-2] IN ROCKIT

        #%% Specifying the objective

        # Minimize moving frame invariants to deal with singularities and noise
        objective_reg = ocp.sum(cas.dot(invars[1:3],invars[1:3]))
        objective = objective_reg/(N-1)
        ocp.add_objective(objective)

        #%% Define solver and save variables
        
        if fatrop_solver:
            ocp.method(rockit.external_method('fatrop', N=N-1))
            ocp._method.set_name("/codegen/rotation") # pick a unique name when using multiple OCP specifications in the same script
            # self.ocp._transcribe()
            # self.ocp._method.set_option("iterative_refinement", False)
            # self.ocp._method.set_option("tol", 1e-8)
        else:
            ocp.method(rockit.MultipleShooting(N=N-1))
            ocp.solver('ipopt', {'expand':True})
            # ocp.solver('ipopt',{"print_time":True,"expand":True},{'gamma_theta':1e-12,'max_iter':200,'tol':1e-4,'print_level':5,'ma57_automatic_scaling':'no','linear_solver':'mumps'})
        
        # Solve already once with dummy values for code generation (TODO: can this step be avoided somehow?)
        ocp.set_initial(R_r_x, np.array([1,0,0]))
        ocp.set_initial(R_r_y, np.array([0,1,0]))
        ocp.set_initial(R_r_z, np.array([0,0,1]))
        ocp.set_initial(invars, np.array([1,0.01,0.01]))
        ocp.set_initial(R_obj_x, np.array([1,0,0]))
        ocp.set_initial(R_obj_y, np.array([0,1,0]))
        ocp.set_initial(R_obj_z, np.array([0,0,1]))
        ocp.set_value(R_obj_m_x, np.tile(np.array([1,0,0]), (N,1)).T)
        ocp.set_value(R_obj_m_y, np.tile(np.array([0,1,0]), (N,1)).T)
        ocp.set_value(R_obj_m_z, np.tile(np.array([0,0,1]), (N,1)).T)
        ocp.set_value(h, 0.01)
        ocp.solve_limited() # code generation

        # Set solver options (TODO: why can this not be done before solving?)
        if fatrop_solver:
            ocp._method.set_option("tol",1e-6)
            #ocp._method.set_option("print_level",0)
        self.first_time = True
        
        # Encapsulate whole rockit specification in a casadi function
        R_obj_m_x_sampled = ocp.sample(R_obj_m_x, grid='control')[1] # sampled measured object orientation (first axis)
        R_obj_m_y_sampled = ocp.sample(R_obj_m_y, grid='control')[1] # sampled measured object orientation (second axis)
        R_obj_m_z_sampled = ocp.sample(R_obj_m_z, grid='control')[1] # sampled measured object orientation (third axis)
        h_value = ocp.value(h) # value of stepsize
        solution = [ocp.sample(invars, grid='control-')[1],
            ocp.sample(R_obj_x, grid='control')[1], # sampled object orientation (first axis)
            ocp.sample(R_obj_y, grid='control')[1], # sampled object orientation (second axis)
            ocp.sample(R_obj_z, grid='control')[1], # sampled object orientation (third axis)
            ocp.sample(R_r_x, grid='control')[1], # sampled FS frame (first axis)
            ocp.sample(R_r_y, grid='control')[1], # sampled FS frame (second axis)
            ocp.sample(R_r_z, grid='control')[1]] # sampled FS frame (third axis)
        
        self.ocp = ocp # save the optimization problem locally, avoids problems when multiple rockit ocp's are created
        self.ocp_function = self.ocp.to_function('ocp_function', 
            [R_obj_m_x_sampled,R_obj_m_y_sampled,R_obj_m_z_sampled,h_value,*solution], # inputs
            [*solution], # outputs
            ["R_obj_m_x","R_obj_m_y","R_obj_m_z","h","invars1","R_obj_x","R_obj_y","R_obj_z","R_r_x1","R_r_y1","R_r_z1"], # input labels for debugging
            ["invars2","R_obj_x2","R_obj_y2","R_obj_z2","R_r_x2","R_r_y2","R_r_z2"], # output labels for debugging
        )

    def calculate_invariants(self,measured_rotations,stepsize): 

        # Check if this is the first function call
        if self.first_time:
            # Initialize states and controls using measurements
            self.solution,measured_rotations = initialize_VI_rot(measured_rotations)
            self.first_time = False

        measured_orientations = [measured_rotations[:,:,0].T,measured_rotations[:,:,1].T,measured_rotations[:,:,2].T]

        # Solve the optimization problem for the given measurements starting from previous solution
        self.solution = self.ocp_function(*measured_orientations, stepsize, *self.solution)

        # Return the results    
        invars, R_obj_x_sol, R_obj_y_sol, R_obj_z_sol, R_r_x_sol, R_r_y_sol, R_r_z_sol,  = self.solution # unpack the results            
        invariants = np.array(invars).T # make a N-1 x 3 array
        invariants = np.vstack((invariants, invariants[-1,:])) # make a N x 3 array by repeating last row
        calculated_trajectory = np.reshape(np.hstack((R_obj_x_sol[:], R_obj_y_sol[:], R_obj_z_sol[:])), (-1,3,3)) # make a N x 3 x 3 array
        calculated_movingframe = np.reshape(np.hstack((R_r_x_sol[:], R_r_y_sol[:], R_r_z_sol[:])), (-1,3,3)) # make a N x 3 x 3 array
        
        return invariants, calculated_trajectory, calculated_movingframe

if __name__ == "__main__":

    # Test data
    measured_orientations = SO3.random_traj(N=3) # TODO replace with something more realistic, now it will sometimes fail
    timestep = 0.1
    
    # Specify OCP symbolically
    N = np.size(measured_orientations,0)
    OCP = OCP_calc_rot(window_len=N,fatrop_solver=False, rms_error_traj=10*pi/180)

    # Solve the OCP using the specified data
    calc_invariants, calc_trajectory, calc_movingframes = OCP.calculate_invariants(measured_orientations, timestep)
    print(calc_invariants)




import numpy as np
import casadi as cas
import invariants_py.dynamics_invariants as dynamics

class OCP_gen_rot:

    def __init__(self, window_len = 100, bool_unsigned_invariants = False, w_pos = 1, w_rot = 1, w_invars = (10**-3)*np.array([1.0, 1.0, 1.0]), max_iters = 300):
       
        #%% Create decision variables and parameters for the optimization problem
        
        opti = cas.Opti() # use OptiStack package from Casadi for easy bookkeeping of variables (no cumbersome indexing)

        # Define system states X (unknown object pose + moving frame pose at every time step) 
        R_obj = []
        R_r = []
        X = []
        for k in range(window_len):
            R_obj.append(opti.variable(3,3)) # object orientation
            R_r.append(opti.variable(3,3)) # rotational Frenet-Serret frame
            X.append(cas.vertcat(cas.vec(R_r[k]), cas.vec(R_obj[k])))

        # Define system controls (invariants at every time step)
        U = opti.variable(3,window_len-1)

        # Define system parameters P (known values in optimization that need to be set right before solving)
        h = opti.parameter(1,1)
        
        # Boundary values
        R_r_start = opti.parameter(3,3)
        R_r_end = opti.parameter(3,3)
        R_obj_start = opti.parameter(3,3)
        R_obj_end = opti.parameter(3,3)
        
        U_demo = opti.parameter(3,window_len-1) # model invariants
        

        #%% Specifying the constraints
        
        # Constrain rotation matrices to be orthogonal (only needed for one timestep, property is propagated by integrator)
        opti.subject_to( R_r[0].T @ R_r[0] == np.eye(3))
        opti.subject_to( R_obj[0].T @ R_obj[0] == np.eye(3))
        
        # Boundary constraints
        opti.subject_to(R_r[0] == R_r_start)
        opti.subject_to(R_r[-1] == R_r_end)
        opti.subject_to(R_obj[0] == R_obj_start)
        opti.subject_to(R_obj[-1] == R_obj_end)
            
        # Dynamic constraints
        integrator = dynamics.define_geom_integrator_rot_FSI_casadi(h)
        for k in range(window_len-1):
            # Integrate current state to obtain next state (next rotation and position)
            Xk_end = integrator(X[k],U[:,k],h)
            
            # Gap closing constraint
            opti.subject_to(Xk_end==X[k+1])
            
        #%% Specifying the objective

        # Fitting constraint to remain close to measurements
        objective_fit = 0
        for k in range(window_len-1):
            err_invars = w_invars*(U[:,k] - U_demo[:,k])
            objective_fit = objective_fit + cas.dot(err_invars,err_invars)
            # objective_fit = objective_fit + 1/window_len*cas.dot(err_invars,err_invars)

        objective = objective_fit

        #%% Define solver and save variables
        opti.minimize(objective)
        opti.solver('ipopt',{"print_time":True,"expand":True},{'tol':1e-4,'print_level':0,'ma57_automatic_scaling':'no','linear_solver':'mumps','max_iter':100})
        
        # Save variables
        self.R_r = R_r
        self.R_obj = R_obj
        self.U = U
        self.U_demo = U_demo
        self.R_r_start = R_r_start
        self.R_r_end = R_r_end
        self.R_obj_start = R_obj_start
        self.R_obj_end = R_obj_end
        self.h = h
        self.window_len = window_len
        self.opti = opti
        
         
    def generate_trajectory(self,U_demo,R_obj_init,R_r_init,R_r_start,R_r_end,R_obj_start,R_obj_end,step_size):
        #%%

        N = self.window_len
        
        # Initialize states
        for k in range(N):
            self.opti.set_initial(self.R_r[k], R_r_init[k])
            self.opti.set_initial(self.R_obj[k], R_obj_init[k])
            
        # Initialize controls
        for k in range(N-1):    
            self.opti.set_initial(self.U[:,k], U_demo[k,:])

        # Set values boundary constraints
        self.opti.set_value(self.R_r_start,R_r_start)
        self.opti.set_value(self.R_r_end,R_r_end)
        self.opti.set_value(self.R_obj_start,R_obj_start)
        self.opti.set_value(self.R_obj_end,R_obj_end)
                
        # Set values parameters
        self.opti.set_value(self.h,step_size)
        for k in range(N-1):
            self.opti.set_value(self.U_demo[:,k], U_demo[k,:])     
        
        # ######################
        # ##  DEBUGGING: check integrator in initial values, time step 0 to 1
        # x0 = cas.vertcat(cas.vec(np.eye(3,3)), cas.vec(measured_positions[0]))
        # u0 = 1e-8*np.ones((3,1))
        # integrator = dynamics.define_geom_integrator_tra_FSI_casadi(self.stepsize)
        # x1 = integrator(x0,u0)
        # print(x1)
        # ######################

        # Solve the NLP
        sol = self.opti.solve_limited()
        self.sol = sol
        
        
        # Extract the solved variables
        invariants = sol.value(self.U).T
        invariants =  np.vstack((invariants,[invariants[-1,:]]))
        calculated_trajectory = np.array([sol.value(i) for i in self.R_obj])
        calculated_movingframe = np.array([sol.value(i) for i in self.R_r])
        
        return invariants, calculated_trajectory, calculated_movingframe



# -*- coding: utf-8 -*-
"""
Created on Wed Mar 6 2024

note: urdf2casadi is not included in the invariants_py installation
install separately using pip install urdf2casadi or include it in the pyproject.toml file

@author: Riccardo
"""

import numpy as np
import casadi as cas
import rockit
import os
from math import pi
import invariants_py.data_handler as dh
import invariants_py.reparameterization as reparam
import urdf2casadi.urdfparser as u2c
from invariants_py.SE3 import rotate_z

def inv_kin(q_init, q_joint_lim, des_p_obj, des_R_obj, window_len = 100, fatrop_solver = False):
    
    # ocp = rockit.Ocp(T=1.0)
    opti = cas.Opti()

    # Define system states
    # q = ocp.state(6) # robot joints
    # p_obj = ocp.state(3) # object position
    # R_obj_x = ocp.state(3,1) # object orientation
    # R_obj_y = ocp.state(3,1) # object orientation
    # R_obj_z = ocp.state(3,1) # object orientation
    # R_obj = cas.horzcat(R_obj_x,R_obj_y,R_obj_z)
    q = opti.variable(6,window_len)

    # Boundary values
    # q_lim = ocp.parameter(6)
    q_lim = opti.parameter(6,1)

    # Parameters
    # p_obj_m = ocp.parameter(3)#,grid='control',include_last=True)
    # R_obj_m_x = ocp.parameter(3,1)#,grid='control',include_last=True)
    # R_obj_m_y = ocp.parameter(3,1)#,grid='control',include_last=True)
    # R_obj_m_z = ocp.parameter(3,1)#,grid='control',include_last=True)
    # R_obj_m = cas.horzcat(R_obj_m_x,R_obj_m_y,R_obj_m_z)
    p_obj_m = opti.parameter(3,window_len)
    R_obj_m = []
    for k in range(window_len):
        R_obj_m.append(opti.parameter(3,3))


    # Boundary constraints
    # ocp.subject_to(-q_lim[0] <= (q[0] <= q_lim[0]))
    # ocp.subject_to(-q_lim[1] <= (q[1] <= q_lim[1]))
    # ocp.subject_to(-q_lim[2] <= (q[2] <= q_lim[2]))
    # ocp.subject_to(-q_lim[3] <= (q[3] <= q_lim[3]))
    # ocp.subject_to(-q_lim[4] <= (q[4] <= q_lim[4]))
    # ocp.subject_to(-q_lim[5] <= (q[5] <= q_lim[5]))
    opti.subject_to(-q_lim[0] <= (q[0] <= q_lim[0]))
    opti.subject_to(-q_lim[1] <= (q[1] <= q_lim[1]))
    opti.subject_to(-q_lim[2] <= (q[2] <= q_lim[2]))
    opti.subject_to(-q_lim[3] <= (q[3] <= q_lim[3]))
    opti.subject_to(-q_lim[4] <= (q[4] <= q_lim[4]))
    opti.subject_to(-q_lim[5] <= (q[5] <= q_lim[5]))

    root = "base_link"
    tip = "TCP_frame"
    ur10 = u2c.URDFparser()
    path_to_urdf = dh.find_data_path("ur10.urdf")
    ur10.from_file(path_to_urdf)
    fk_dict = ur10.get_forward_kinematics(root, tip)
    forward_kinematics = fk_dict["T_fk"]
    objective = 0
    for k in range(window_len):
        T_rob = forward_kinematics(q[:,k].T)
        T_rob = rotate_z(pi) @ T_rob
        p_obj = T_rob[:3,3]
        R_obj = T_rob[:3,:3]
        # Specify the objective
        e_pos = cas.dot(p_obj - p_obj_m[:,k],p_obj - p_obj_m[:,k])
        e_rot = cas.dot(R_obj_m[k].T @ R_obj - np.eye(3),R_obj_m[k].T @ R_obj - np.eye(3))
        # objective = ocp.sum(e_pos + e_rot)
        if k == 0:
            objective = objective + e_pos + e_rot
        else:
            qdot = q[:,k] - q[:,k-1]
            objective = objective + e_pos + e_rot + 0.001*cas.dot(qdot,qdot)
    # ocp.add_objective(objective)
    opti.minimize(objective)

    if fatrop_solver:
        # ocp.method(rockit.external_method('fatrop' , N=window_len-1))
        pass
    else:
        # ocp.method(rockit.MultipleShooting(N=window_len-1))
        # ocp.solver('ipopt', {'expand':True})
        opti.solver('ipopt',{"print_time":True,"expand":True},{'gamma_theta':1e-12,'tol':1e-4,'print_level':5,'ma57_automatic_scaling':'no','linear_solver':'mumps','max_iter':100})

    # ocp.set_initial(q,q_init)
    # ocp.set_initial(p_obj,p_obj_init)
    # ocp.set_initial(R_obj_x,R_obj_init[:,:,0])
    # ocp.set_initial(R_obj_y,R_obj_init[:,:,1])
    # ocp.set_initial(R_obj_z,R_obj_init[:,:,2])
    for k in range(window_len):
        opti.set_initial(q[:,k],q_init[k].T)

    # Set values of boundary constraints
    # ocp.set_value(q_lim,q_joint_lim)
    opti.set_value(q_lim,q_joint_lim)

    #Set values of parameters
    # ocp.set_value(p_obj_m,des_p_obj)
    # ocp.set_value(R_obj_m_x,des_R_obj[:,0])
    # ocp.set_value(R_obj_m_y,des_R_obj[:,1])
    # ocp.set_value(R_obj_m_z,des_R_obj[:,2])
    for k in range(window_len):
        opti.set_value(p_obj_m[:,k],des_p_obj[k].T)
        opti.set_value(R_obj_m[k],des_R_obj[k])

    # Solve the NLP
    # sol = ocp.solve()
    sol = opti.solve_limited()

    joint_val = sol.value(q)

    return joint_val.T

 
if __name__ == "__main__":

    data_location = dh.find_data_path('beer_1.txt')
    trajectory,time = dh.read_pose_trajectory_from_txt(data_location)
    pose,time_profile,arclength,nb_samples,stepsize = reparam.reparameterize_trajectory_arclength(trajectory)
    startpos = [0.3056, 0.0635, 0.441]
    des_p_obj = pose[:,:3,3]  + startpos
    des_R_obj = pose[:,:3,:3]
    q_init = [-pi, -2.27, 2.27, -pi/2, -pi/2, pi/4] * np.ones((100,6))
    # q_des = [-3.18, -1.19, 1.37, -1.74, -1.58, 0.75]
    # p_obj_init = [0.9107, 0.03, 0.435]
    # R_obj_init = np.eye(3)
    q_joint_lim = [2*pi, 2*pi, pi, 2*pi, 2*pi, 2*pi]

    q = inv_kin(q_init, q_joint_lim, des_p_obj, des_R_obj)
    print(q)


"""
    This module contains implmentations of the geometries
    which can be used to model the objects for monitoring collisions

    Contains class for convex polytopes

    Other geometries can be added as well and be used with the GJK algorithm

    Script taken from: https://github.com/mohith-sakthivel/Fast_GJK-collision_detection_bottle
"""
import numpy as np
# from rotation_ops import RotatePos
# np.random.seed(0)


class Polytope():
    def __init__(self, PointList):
        """
        Create and intialize a convex ploytope at the robot
        end effector
        """
        self.Vertices = np.array(PointList, dtype='float64')
        assert len(self.Vertices.shape) == 2, \
            "Atleast 1 point must be specified.\n" +\
            "All points must be in 3D space"
        assert self.Vertices.shape[1] == 3, "All points must be in 3D space"
        self.NoOfVertices = len(PointList)
        self.CurrentVertices = np.copy(self.Vertices)

    # def UpdatePosition(self, pos, quat):
    #     """
    #     Updates the vertices of the polytope based on the
    #     provided position vector and quaternion
    #     """
    #     pos = np.array(pos, dtype="float64")
    #     quat = np.array(quat, dtype="float64")
    #     for i in range(len(self.Vertices)):
    #         self.CurrentVertices[i] = pos+RotatePos(self.Vertices[i], quat)

    def SupportFunc(self, dVector):
        """
        Provides the farthest point in the polytope along the
        provided direction dVector

        Inputs:
            dVector      - direction vector in 3D space
        Output:
            Farthest point in polytope along the direction dVector,
            Projection of the farthest point on the direction vector
        """
        MaxDist = np.dot(dVector, self.CurrentVertices[0])
        VertexIndex = 0
        for i in range(1, len(self.CurrentVertices)):
            dist = np.dot(dVector, self.CurrentVertices[i])
            if dist > MaxDist:
                MaxDist = dist
                VertexIndex = i
        return np.array([self.CurrentVertices[VertexIndex],
                         MaxDist, VertexIndex], dtype=object)



import numpy as np
import casadi as cas
import invariants_py.dynamics_invariants as dynamics
from invariants_py.ocp_helper import jerk_invariant

class OCP_calc_pos:

    def __init__(self, window_len = 100, bool_unsigned_invariants = False, w_pos = 1, w_regul = (10**-6)*np.array([1.0, 1.0, 1.0]), planar_task = False):

        # System states
        R_t  = cas.MX.sym('R_t',3,3) # translational Frenet-Serret frame
        p_obj = cas.MX.sym('p_obj',3,1) # object position
        i1dot = cas.MX.sym('i1dot',1,1)
        i1 = cas.MX.sym('i1',1,1)
        i2 = cas.MX.sym('i2',1,1)
        x = cas.vertcat(cas.vec(R_t), p_obj, i1dot, i1, i2)

        #np = length(R_obj(:)) + length(p_obj)

        # System controls (invariants)
        #u = cas.MX.sym('i',3,1)
        i1ddot = cas.MX.sym('i1ddot')
        i2dot = cas.MX.sym('i2dot')
        i3 = cas.MX.sym('i3')
        u = cas.vertcat(i1ddot,i2dot,i3)

        opti = cas.Opti() # use OptiStack package from Casadi for easy bookkeeping of variables (no cumbersome indexing)
        h = opti.parameter(1,1)
       
        #%% Define geometric integrator
        ## Define a geometric integrator for eFSI, (meaning rigid-body motion is perfectly integrated assuming constant invariants)
        invariants = cas.vertcat(i1,i2,i3)
        (R_t_plus1, p_obj_plus1) = dynamics.vector_invariants_position(R_t, p_obj, invariants, h)

        i1dotplus1 = i1dot + i1ddot * h
        i1plus1 = i1 + i1dot * h + i1ddot * h**2/2
        i2plus1 = i2 + i2dot * h

        out_plus1 = cas.vertcat(cas.vec(R_t_plus1),p_obj_plus1,i1dotplus1,i1plus1,i2plus1)

        integrator = cas.Function("phi", [x,u,h] , [out_plus1])


       
        #%% Create decision variables and parameters for the optimization problem
        
        # Define system states X (unknown object pose + moving frame pose at every time step) 
        p_obj = []
        R_t = []
        i1dot = []
        i1 = []
        i2 = []
        X = []
        for k in range(window_len):
            p_obj.append(opti.variable(3,1)) # object position
            R_t.append(opti.variable(3,3)) # translational Frenet-Serret frame
            i1dot.append(opti.variable(1,1))
            i1.append(opti.variable(1,1))
            i2.append(opti.variable(1,1))
            
            X.append(cas.vertcat(cas.vec(R_t[k]), cas.vec(p_obj[k]), cas.vec(i1dot[k]), cas.vec(i1[k]), cas.vec(i2[k]) ))

        # Define system controls (invariants at every time step)
        i1ddot = opti.variable(1,window_len-1)
        i2dot = opti.variable(1,window_len-1)
        i3 = opti.variable(1,window_len-1)
        U = cas.vertcat(i1ddot,i2dot,i3)

        # Define system parameters P (known values in optimization that need to be set right before solving)
        p_obj_m = [] # measured object positions
        R_t_0 = opti.parameter(3,3) # initial translational Frenet-Serret frame at first sample of window
        for k in range(window_len):
            p_obj_m.append(opti.parameter(3,1)) # object position
  
        #%% Specifying the constraints
        
        # Constrain rotation matrices to be orthogonal (only needed for one timestep, property is propagated by integrator)
        #áopti.subject_to( R_t[-1].T @ R_t[-1] == np.eye(3))
        constr = R_t[0].T @ R_t[0]
        opti.subject_to(cas.vertcat(constr[1,0], constr[2,0], constr[2,1], constr[0,0], constr[1,1], constr[2,2])==cas.vertcat(cas.DM.zeros(3,1), cas.DM.ones(3,1)))


        # Dynamic constraints
        for k in range(window_len-1):
            # Integrate current state to obtain next state (next rotation and position)
            Xk_end = integrator(X[k],U[:,k],h)
            
            # Gap closing constraint
            opti.subject_to(Xk_end==X[k+1])
           
            
        # 2D contour   
        if planar_task:
            for k in range(window_len):
                opti.subject_to( cas.dot(R_t[k][:,2],np.array([0,0,1])) > 0)
            
            
        # Lower bounds on controls
        #if bool_unsigned_invariants:
        #    opti.subject_to(U[0,:]>=0) # lower bounds on control
        #    opti.subject_to(U[1,:]>=0) # lower bounds on control

        #%% Specifying the objective

        # Fitting constraint to remain close to measurements
        objective_fit = 0
        for k in range(window_len):
            err_pos = p_obj[k] - p_obj_m[k] # position error
            objective_fit = objective_fit + 1/window_len* w_pos*cas.dot(err_pos,err_pos)

        # Regularization constraints to deal with singularities and noise
        objective_reg = 0
        for k in range(window_len-1):

            jk = jerk_invariant(i1[k],i1dot[k],i1ddot[k],i2[k],i2dot[k],i3[k])

            #if k!=0:
            #    err_deriv = U[:,k] - U[:,k-1] # first-order finite backwards derivative (noise smoothing effect)
            #else:
            #    err_deriv = 0
            err_abs = 10e-6*U[[1,2],k] # absolute value invariants (force arbitrary invariants to zero)

            jk_weighted = w_regul**(0.5)*jk

            objective_reg = objective_reg + 1/window_len * cas.dot(jk_weighted,jk_weighted)

            ##Check that obj function is correctly typed in !!!
            #objective_reg = objective_reg \
            #                + cas.dot(w_deriv**(0.5)*err_deriv,w_deriv**(0.5)*err_deriv) \
            #                + cas.dot(w_abs**(0.5)*err_abs, w_abs**(0.5)*err_abs)

        objective = objective_fit + objective_reg + cas.dot(err_abs,err_abs)

        #%% Define solver and save variables
        opti.minimize(objective)
        opti.solver('ipopt',{"print_time":False,"expand":False},{'max_iter':200,'tol':1e-8,'print_level':0,'ma57_automatic_scaling':'no','linear_solver':'mumps'})

        
        # Save variables
        self.R_t = R_t
        self.p_obj = p_obj
        self.i1dot = i1dot
        self.i1 = i1
        self.i2 = i2
        self.i1ddot = i1ddot
        self.i2dot = i2dot
        self.i3 = i3
        self.p_obj_m = p_obj_m
        self.R_t_0 = R_t_0
        self.window_len = window_len
        self.opti = opti
        self.first_window = True
        self.h = h
    
        self.dummy_solve()
    
    
    def dummy_solve(self):
        print("dummy solve")
        for k in range(self.window_len):
            self.opti.set_initial(self.R_t[k], np.eye(3) )
            self.opti.set_initial(self.i1[k],1e-12)
            self.opti.set_initial(self.i2[k],1e-12)
            self.opti.set_value(self.p_obj_m[k], np.zeros((1,3)))
        self.opti.set_value(self.h,1)
        self.opti.set_value(self.R_t_0, np.eye(3,3))
        self.opti.solve_limited()
    
    def calculate_invariants_global(self,trajectory_meas,stepsize):
        #%%

        if trajectory_meas.shape[1] == 3:
            measured_positions = trajectory_meas
        else:
            measured_positions = trajectory_meas[:,:3,3]
        N = self.window_len
        
        # Estimate initial moving frames based on measurements
        #TODO  this is not correct yet, ex not perpendicular to ey + not robust for singularities, these parts must still be transferred from Matlab
        Pdiff = np.diff(measured_positions,axis=0)
        ex = Pdiff / np.linalg.norm(Pdiff,axis=1).reshape(N-1,1)
        ex = np.vstack((ex,[ex[-1,:]]))
        ez = np.tile( np.array((0,0,1)), (N,1) )
        ey = np.array([np.cross(ez[i,:],ex[i,:]) for i in range(N)])
        
        # Initialize states
        for k in range(N):
            self.opti.set_initial(self.R_t[k], np.array([ex[k,:], ey[k,:], ez[k,:]]).T ) #construct_init_FS_from_traj(meas_traj.Obj_location)
            self.opti.set_initial(self.p_obj[k], measured_positions[k])
            self.opti.set_initial(self.i1dot[k],1e-12)
            self.opti.set_initial(self.i1[k],1e0)
            self.opti.set_initial(self.i2[k],1e-5)
            
        # Initialize controls
        for k in range(N-1):    
            self.opti.set_initial(self.i1ddot[k],1e-12)
            self.opti.set_initial(self.i2dot[k],1e-12)
            self.opti.set_initial(self.i3[k],1e-12)
            #self.opti.set_initial(self.U[:,k], 1e-3*np.ones((3,1)))
            
        # Set values parameters
        self.opti.set_value(self.R_t_0, np.eye(3,3))
        for k in range(N):
            self.opti.set_value(self.p_obj_m[k], measured_positions[k])       
        
        self.opti.set_value(self.h,stepsize)

        # ######################
        # ##  DEBUGGING: check integrator in initial values, time step 0 to 1
        # x0 = cas.vertcat(cas.vec(np.eye(3,3)), cas.vec(measured_positions[0]))
        # u0 = 1e-8*np.ones((3,1))
        # integrator = dynamics.define_geom_integrator_tra_FSI_casadi(self.stepsize)
        # x1 = integrator(x0,u0)
        # print(x1)
        # ######################

        # Solve the NLP
        sol = self.opti.solve_limited()
        self.sol = sol
        
        # Extract the solved variables
  
        i1 = np.array([sol.value(i) for i in self.i1])
        i2 = np.array([sol.value(i) for i in self.i2])
        i3 = sol.value(self.i3)
        i3 = np.append(i3,i3[-1])
        invariants = np.array((i1,i2,i3)).T
        
        calculated_trajectory = np.array([sol.value(i) for i in self.p_obj])
        calculated_movingframe = np.array([sol.value(i) for i in self.R_t])
        
        return invariants, calculated_trajectory, calculated_movingframe

    def calculate_invariants_online(self,trajectory_meas,stepsize,sample_jump):
        #%%
        if self.first_window:
            # Calculate invariants in first window
            invariants, calculated_trajectory, calculated_movingframe = self.calculate_invariants_global(trajectory_meas,stepsize)
            self.first_window = False
                    
            # Add continuity constraints on first sample
            #self.opti.subject_to( self.R_t[0] == self.R_t_0 )
            #self.opti.subject_to( self.p_obj[0] == self.p_obj_m[0])
            
            return invariants, calculated_trajectory, calculated_movingframe
        else:
            
            if trajectory_meas.shape[1] == 3:
                measured_positions = trajectory_meas
            else:
                measured_positions = trajectory_meas[:,:3,3]
            N = self.window_len
            
            #%% Set values parameters
            #for k in range(1,N):
            #    self.opti.set_value(self.p_obj_m[k], measured_positions[k-1])   
            
            for k in range(0,N):
                    self.opti.set_value(self.p_obj_m[k], measured_positions[k])   
            
            # Set other parameters equal to the measurements in that window
            self.opti.set_value(self.R_t_0, self.sol.value(self.R_t[sample_jump]));
            #self.opti.set_value(self.p_obj_m[0], self.sol.value(self.p_obj[sample_jump]));
            
            self.opti.set_value(self.h,stepsize);
        
            #%% First part of window initialized using results from earlier solution
            # Initialize states
            for k in range(N-sample_jump-1):
                self.opti.set_initial(self.R_t[k], self.sol.value(self.R_t[sample_jump+k]))
                self.opti.set_initial(self.p_obj[k], self.sol.value(self.p_obj[sample_jump+k]))
                self.opti.set_initial(self.i1dot[k], self.sol.value(self.i1dot[sample_jump+k]))
                self.opti.set_initial(self.i1[k], self.sol.value(self.i1[sample_jump+k]))
                self.opti.set_initial(self.i2[k], self.sol.value(self.i2[sample_jump+k]))
                
            # Initialize controls
            for k in range(N-sample_jump-1):    
                #○self.opti.set_initial(self.U[:,k], self.sol.value(self.U[:,sample_jump+k]));
                self.opti.set_initial(self.i1ddot[k], self.sol.value(self.i1ddot[:,sample_jump+k]))
                self.opti.set_initial(self.i2dot[k], self.sol.value(self.i2dot[:,sample_jump+k]))
                self.opti.set_initial(self.i3[k], self.sol.value(self.i3[:,sample_jump+k]))
                
            #%% Second part of window initialized uses default initialization
            # Initialize states
            for k in range(N-sample_jump,N):
                self.opti.set_initial(self.R_t[k], self.sol.value(self.R_t[N-5]))
                self.opti.set_initial(self.p_obj[k], measured_positions[k-1])
                self.opti.set_initial(self.i1dot[k],1e-12)
                self.opti.set_initial(self.i1[k],1e0)
                self.opti.set_initial(self.i2[k],1e-5)
                
                
            # Initialize controls
            for k in range(N-sample_jump-1,N-1):    
                #self.opti.set_initial(self.U[:,k], 1e-3*np.ones((3,1)))
                self.opti.set_initial(self.i1ddot[k],1e-12)
                self.opti.set_initial(self.i2dot[k],1e-12)
                self.opti.set_initial(self.i3[k],1e-12)
                

            #print(self.sol.value(self.R_t[-1]))

            #%% Solve the NLP
            sol = self.opti.solve_limited()
            self.sol = sol
            
            # Extract the solved variables

            i1 = np.array([sol.value(i) for i in self.i1])
            i2 = np.array([sol.value(i) for i in self.i2])
            i3 = sol.value(self.i3)
            i3 = np.append(i3,i3[-1])
            invariants = np.array((i1,i2,i3)).T
            
            calculated_trajectory = np.array([sol.value(i) for i in self.p_obj])
            calculated_movingframe = np.array([sol.value(i) for i in self.R_t])
            
            return invariants, calculated_trajectory, calculated_movingframe



import numpy as np
import casadi as cas
import rockit
import invariants_py.dynamics_invariants as dynamics
import time
from invariants_py.ocp_helper import check_solver, tril_vec, tril_vec_no_diag, diffR, diag
from invariants_py.SO3 import rotate_x
from invariants_py.initialization import generate_initvals_from_bounds_rot

class OCP_gen_rot:


    def __init__(self, boundary_constraints, window_len = 100, bool_unsigned_invariants = False, w_pos = 1, w_rot = 1, max_iters = 300, fatrop_solver = False):
        
        fatrop_solver = check_solver(fatrop_solver)               
       
        #%% Create decision variables and parameters for the optimization problem
        
        ocp = rockit.Ocp(T=1.0)

        # Define system states X (unknown object pose + moving frame pose at every time step) 
        R_obj_x = ocp.state(3,1) # object orientation
        R_obj_y = ocp.state(3,1) # object orientation
        R_obj_z = ocp.state(3,1) # object orientation
        R_obj = cas.horzcat(R_obj_x,R_obj_y,R_obj_z)
        R_r_x = ocp.state(3,1) # rotational Frenet-Serret frame
        R_r_y = ocp.state(3,1) # rotational Frenet-Serret frame
        R_r_z = ocp.state(3,1) # rotational Frenet-Serret frame
        R_r = cas.horzcat(R_r_x,R_r_y,R_r_z)

        # Define system controls (invariants at every time step)
        invars = ocp.control(3)

        # Define system parameters P (known values in optimization that need to be set right before solving)
        h = ocp.parameter(1)
        
        # Boundary values
        if "moving-frame-orientation" in boundary_constraints and "initial" in boundary_constraints["moving-frame-orientation"]:
            R_r_start = ocp.parameter(3,3)
        if "moving-frame-orientation" in boundary_constraints and "final" in boundary_constraints["moving-frame-orientation"]:
            R_r_end = ocp.parameter(3,3)
        if "orientation" in boundary_constraints and "initial" in boundary_constraints["orientation"]:
            R_obj_start = ocp.parameter(3,3)
        if "orientation" in boundary_constraints and "final" in boundary_constraints["orientation"]:
            R_obj_end = ocp.parameter(3,3)
        
        invars_demo = ocp.parameter(3,grid='control',include_last=True) # model invariants
        
        w_invars = ocp.parameter(3,grid='control',include_last=True) # weights for invariants

        #%% Specifying the constraints
        
        # Constrain rotation matrices to be orthogonal (only needed for one timestep, property is propagated by integrator)
        ocp.subject_to(ocp.at_t0(tril_vec(R_r.T @ R_r - np.eye(3))==0.))
        ocp.subject_to(ocp.at_t0(tril_vec(R_obj.T @ R_obj - np.eye(3))==0.))
        
        # Boundary constraints
        if "moving-frame-orientation" in boundary_constraints and "initial" in boundary_constraints["moving-frame-orientation"]:
            ocp.subject_to(ocp.at_t0(tril_vec_no_diag(R_r.T @ R_r_start - np.eye(3)) == 0.))
        if "moving-frame-orientation" in boundary_constraints and "final" in boundary_constraints["moving-frame-orientation"]:
            ocp.subject_to(ocp.at_tf(tril_vec_no_diag(R_r.T @ R_r_end - np.eye(3)) == 0.))
        if "orientation" in boundary_constraints and "initial" in boundary_constraints["orientation"]:
            ocp.subject_to(ocp.at_t0(tril_vec_no_diag(R_obj.T @ R_obj_start - np.eye(3)) == 0.))
        if "orientation" in boundary_constraints and "initial" in boundary_constraints["orientation"]:
            ocp.subject_to(ocp.at_tf(tril_vec_no_diag(R_obj.T @ R_obj_end - np.eye(3)) == 0.))

        #ocp.subject_to(ocp.at_t0(R_r == R_r_start))
        #ocp.subject_to(ocp.at_t0(self.diffR(R_r,R_r_start)) == 0)
        #ocp.subject_to(ocp.at_tf(self.three_elements(R_r) == self.three_elements(R_r_end)))
        #ocp.subject_to(ocp.at_tf(self.diffR(R_r,R_r_end)) == 0)
        #ocp.subject_to(ocp.at_t0(R_obj == R_obj_start))
        #ocp.subject_to(ocp.at_t0(self.diffR(R_obj,R_obj_start)) == 0)
        #ocp.subject_to(ocp.at_tf(self.three_elements(R_obj) == self.three_elements(R_obj_end)))
        #ocp.subject_to(ocp.at_tf(self.three_elements(R_obj) == self.three_elements(R_obj_end)))
        #ocp.subject_to(ocp.at_tf(self.diffR(R_obj,R_obj_end)) == 0)
            
        # Dynamic constraints
        (R_r_plus1, R_obj_plus1) = dynamics.dyn_vector_invariants_rotation(R_r, R_obj, invars, h)
        # Integrate current state to obtain next state (next rotation and position)
        ocp.set_next(R_obj_x,R_obj_plus1[:,0])
        ocp.set_next(R_obj_y,R_obj_plus1[:,1])
        ocp.set_next(R_obj_z,R_obj_plus1[:,2])
        ocp.set_next(R_r_x,R_r_plus1[:,0])
        ocp.set_next(R_r_y,R_r_plus1[:,1])
        ocp.set_next(R_r_z,R_r_plus1[:,2])
            
        #%% Specifying the objective

        # Fitting constraint to remain close to measurements
        objective = ocp.sum(1/window_len*cas.dot(w_invars*(invars - invars_demo),w_invars*(invars - invars_demo)),include_last=True)

        #%% Define solver and save variables
        ocp.add_objective(objective)
        if fatrop_solver:
            ocp.method(rockit.external_method('fatrop' , N=window_len-1))
            # ocp._method.set_name("generation_rotation")            
            ocp._method.set_name("/codegen/generation_rotation")
        else:
            ocp.method(rockit.MultipleShooting(N=window_len-1))
            ocp.solver('ipopt', {'expand':True, 'ipopt.print_info_string': 'yes'})
            # ocp.solver('ipopt',{"print_time":True,"expand":True},{'tol':1e-4,'print_level':0,'ma57_automatic_scaling':'no','linear_solver':'mumps','max_iter':100})
        
        # Solve already once with dummy values for code generation (TODO: can this step be avoided somehow?)
        ocp.set_initial(R_r_x, np.array([1,0,0]))
        ocp.set_initial(R_r_y, np.array([0,1,0]))
        ocp.set_initial(R_r_z, np.array([0,0,1]))
        ocp.set_initial(invars, np.array([1,0.01,0.01]))
        ocp.set_initial(R_obj_x, np.array([1,0,0]))
        ocp.set_initial(R_obj_y, np.array([0,1,0]))
        ocp.set_initial(R_obj_z, np.array([0,0,1]))
        ocp.set_value(invars_demo, 0.001+np.zeros((3,window_len)))
        ocp.set_value(w_invars, 0.001+np.zeros((3,window_len)))
        ocp.set_value(h, 0.01)
        if "moving-frame-orientation" in boundary_constraints and "initial" in boundary_constraints["moving-frame-orientation"]:
            ocp.set_value(R_r_start, np.eye(3))
        if "moving-frame-orientation" in boundary_constraints and "final" in boundary_constraints["moving-frame-orientation"]:
            ocp.set_value(R_r_end, np.eye(3))
        if "orientation" in boundary_constraints and "initial" in boundary_constraints["orientation"]:
            ocp.set_value(R_obj_start, np.eye(3))
        if "orientation" in boundary_constraints and "final" in boundary_constraints["orientation"]:
            ocp.set_value(R_obj_end, rotate_x(np.pi/4))
        ocp.solve_limited() # code generation
        if fatrop_solver:
            tot_time = ocp._method.myOCP.get_stats().time_total
        else:
            tot_time = 0

        self.first_window = True

        # Encapsulate whole rockit specification in a casadi function
        invars_sampled = ocp.sample(invars_demo, grid='control')[1] # sampled demonstration invariants
        w_sampled = ocp.sample(w_invars, grid='control')[1] # sampled invariants weights 
        h_value = ocp.value(h) # value of stepsize

        bounds = []
        bounds_labels = []
        if "orientation" in boundary_constraints and "initial" in boundary_constraints["orientation"]:
            bounds.append(ocp.value(R_obj_start))
            bounds_labels.append("R_obj_start")
        if "orientation" in boundary_constraints and "final" in boundary_constraints["orientation"]:
            bounds.append(ocp.value(R_obj_end))
            bounds_labels.append("R_obj_end")
        if "moving-frame-orientation" in boundary_constraints and "initial" in boundary_constraints["moving-frame-orientation"]:
            bounds.append(ocp.value(R_r_start))
            bounds_labels.append("R_r_start")
        if "moving-frame-orientation" in boundary_constraints and "final" in boundary_constraints["moving-frame-orientation"]:
            bounds.append(ocp.value(R_r_end))
            bounds_labels.append("R_r_end")
        
        solution = [ocp.sample(invars, grid='control-')[1],
            ocp.sample(R_r_x, grid='control')[1], # sampled FS frame (first axis)
            ocp.sample(R_r_y, grid='control')[1], # sampled FS frame (second axis)
            ocp.sample(R_r_z, grid='control')[1], # sampled FS frame (third axis)
            ocp.sample(R_obj_x, grid='control')[1], # sampled object orientation (first axis)
            ocp.sample(R_obj_y, grid='control')[1], # sampled object orientation (second axis)
            ocp.sample(R_obj_z, grid='control')[1]] # sampled object orientation (third axis)

        self.ocp = ocp # save the optimization problem locally, avoids problems when multiple rockit ocp's are created

        self.ocp_function = self.ocp.to_function('ocp_function', 
            [invars_sampled,w_sampled,h_value,*bounds,*solution], # inputs
            [*solution], # outputs
            ["invars","w_invars","stepsize",*bounds_labels,"invars1","R_r_x1","R_r_y1","R_r_z1","R_obj_x1","R_obj_y1","R_obj_z1"], # input labels for debugging
            ["invars2","R_r_x2","R_r_y2","R_r_z2","R_obj_x2","R_obj_y2","R_obj_z2"], # output labels for debugging
        )

        # Save variables
        self.R_r_x = R_r_x
        self.R_r_y = R_r_y
        self.R_r_z = R_r_z
        self.R_r = R_r
        self.R_obj_x = R_obj_x
        self.R_obj_y = R_obj_y
        self.R_obj_z = R_obj_z
        self.R_obj = R_obj
        self.invars = invars
        self.invars_demo = invars_demo
        self.w_invars = w_invars
        if "moving-frame-orientation" in boundary_constraints and "initial" in boundary_constraints["moving-frame-orientation"]:
            self.R_r_start = R_r_start
        if "moving-frame-orientation" in boundary_constraints and "final" in boundary_constraints["moving-frame-orientation"]:
            self.R_r_end = R_r_end
        if "orientation" in boundary_constraints and "initial" in boundary_constraints["orientation"]:
            self.R_obj_start = R_obj_start
        if "orientation" in boundary_constraints and "final" in boundary_constraints["orientation"]:
            self.R_obj_end = R_obj_end
        self.h = h
        self.window_len = window_len
        self.ocp = ocp
        self.sol = None
        self.first_window = True
        self.fatrop = fatrop_solver
        self.tot_time = tot_time
        
         
    def generate_trajectory_OLD(self,invariant_model,boundary_constraints, step_size, weights_params,initial_values):

        N = np.size(invariant_model,0)
        
        R_obj_init = initial_values['trajectory-orientation']
        R_r_init = initial_values['moving-frame-orientation']
        invars_init = initial_values['invariants-orientation']

        R_r_start = boundary_constraints["moving-frame-orientation"]["initial"]
        R_r_end = boundary_constraints["moving-frame-orientation"]["final"]
        R_obj_start = boundary_constraints["orientation"]["initial"]
        R_obj_end = boundary_constraints["orientation"]["final"]

        # Get the weights for the invariants or set default values
        w_invars = weights_params.get('w_invars', (10**-3)*np.array([1.0, 1.0, 1.0]))
        w_high_start = weights_params.get('w_high_start', 1)
        w_high_end = weights_params.get('w_high_end', 0)
        w_high_invars = weights_params.get('w_high_invars', (10**-3)*np.array([1.0, 1.0, 1.0]))
        w_high_active = weights_params.get('w_high_active', 0)

        # Set the weights for the invariants
        weights = np.tile(w_invars, (len(invariant_model), 1))
        if w_high_active:
            weights[w_high_start:w_high_end+1, :] = w_high_invars
        self.ocp.set_value(self.w_invars, weights.T) 

        # Initialize states
        self.ocp.set_initial(self.R_obj_x, R_obj_init[:self.window_len,:,0].T) 
        self.ocp.set_initial(self.R_obj_y, R_obj_init[:self.window_len,:,1].T) 
        self.ocp.set_initial(self.R_obj_z, R_obj_init[:self.window_len,:,2].T) 
        self.ocp.set_initial(self.R_r_x, R_r_init[:self.window_len,:,0].T) 
        self.ocp.set_initial(self.R_r_y, R_r_init[:self.window_len,:,1].T) 
        self.ocp.set_initial(self.R_r_z, R_r_init[:self.window_len,:,2].T) 
            
        # Initialize controls
        self.ocp.set_initial(self.invars,invars_init.T)

        # Set values boundary constraints
        self.ocp.set_value(self.R_r_start,R_r_start)
        self.ocp.set_value(self.R_r_end,R_r_end)
        self.ocp.set_value(self.R_obj_start,R_obj_start)
        self.ocp.set_value(self.R_obj_end,R_obj_end)

        # Set values parameters
        self.ocp.set_value(self.h,step_size)
        self.ocp.set_value(self.invars_demo, invariant_model.T)  

        # Solve the NLP
        sol = self.ocp.solve()
        # if self.fatrop:
        #     tot_time = self.ocp._method.myOCP.get_stats().time_total
        # else:
        #     tot_time = 0
                      
        # Extract the solved variables
        _,i_r1 = sol.sample(self.invars[0],grid='control')
        _,i_r2 = sol.sample(self.invars[1],grid='control')
        _,i_r3 = sol.sample(self.invars[2],grid='control')
        invariants = np.array((i_r1,i_r2,i_r3)).T
        _,calculated_trajectory = sol.sample(self.R_obj,grid='control')
        _,calculated_movingframe = sol.sample(self.R_r,grid='control')
                
        return invariants, calculated_trajectory, calculated_movingframe, self.tot_time
    
    def generate_trajectory(self, invariant_model, boundary_constraints, step_size, weights_params = {}, initial_values = {}):

        N = invariant_model.shape[0]

        # Get the weights for the invariants or set default values
        w_invars = weights_params.get('w_invars', (10**-3)*np.array([1.0, 1.0, 1.0]))
        w_high_start = weights_params.get('w_high_start', 1)
        w_high_end = weights_params.get('w_high_end', 0)
        w_high_invars = weights_params.get('w_high_invars', (10**-3)*np.array([1.0, 1.0, 1.0]))
        w_high_active = weights_params.get('w_high_active', 0)

        # Set the weights for the invariants
        w_invars = np.tile(w_invars, (len(invariant_model),1)).T
        if w_high_active:
            w_invars[:, w_high_start:w_high_end+1] = w_high_invars.reshape(-1, 1)

        boundary_values_list = [value for sublist in boundary_constraints.values() for value in sublist.values()]

        if self.first_window and not initial_values:
            self.solution = generate_initvals_from_bounds_rot(boundary_constraints, np.size(invariant_model,0))
            self.first_window = False
        elif self.first_window:
            self.solution = [initial_values["invariants-orientation"][:N-1,:].T, initial_values["moving-frame-orientation"][:N,:,0].T, initial_values["moving-frame-orientation"][:N,:,1].T, initial_values["moving-frame-orientation"][:N,:,2].T, initial_values["trajectory-orientation"][:N,:,0].T, initial_values["trajectory-orientation"][:N,:,1].T, initial_values["trajectory-orientation"][:N,:,2].T]
            self.first_window = False


        # Call solve function
        self.solution = self.ocp_function(invariant_model.T,w_invars,step_size,*boundary_values_list,*self.solution)

        #Return the results
        invars_sol, R_r_x_sol, R_r_y_sol, R_r_z_sol, R_obj_x_sol, R_obj_y_sol, R_obj_z_sol = self.solution # unpack the results            
        invariants = np.array(invars_sol).T
        invariants = np.vstack((invariants, invariants[-1,:])) # make a N x 3 array by repeating last row
        calculated_trajectory = np.reshape(np.hstack((R_obj_x_sol[:], R_obj_y_sol[:], R_obj_z_sol[:])), (-1,3,3)) # make a N x 3 x 3 array
        calculated_movingframe = np.reshape(np.hstack((R_r_x_sol[:], R_r_y_sol[:], R_r_z_sol[:])), (-1,3,3)) # make a N x 3 x 3 array

        return invariants, calculated_trajectory, calculated_movingframe, self.tot_time
    
if __name__ == "__main__":

    # Randomly chosen data
    N = 100
    invariant_model = np.ones((N,3))*[1,0.001,0.001]

    # Boundary constraints
    boundary_constraints = {
        "orientation": {
            "initial": np.eye(3),
            "final": rotate_x(np.pi/6)
        },
        "moving-frame-orientation": {
            "initial": np.eye(3),
            "final": np.eye(3)
        },
    }
    step_size = 0.1

    # Create an instance of OCP_gen_pos
    ocp = OCP_gen_rot(boundary_constraints,fatrop_solver=True, window_len=N)

    # Call the generate_trajectory function
    invariants, calculated_trajectory, calculated_movingframe, tot_time = ocp.generate_trajectory(invariant_model, boundary_constraints, step_size)

    # Print the results
    print("Invariants_start:", invariants[0])
    print("Invariants_end:", invariants[-1])
    print("Calculated Trajectory_start:", calculated_trajectory[0])
    print("Calculated Trajectory_end:", calculated_trajectory[-1])
    print("Calculated Moving Frame_start:", calculated_movingframe[0])
    print("Calculated Moving Frame_end:", calculated_movingframe[-1])
    err1 = calculated_trajectory[-1].T @ boundary_constraints["orientation"]["final"]

    # Second call to generate_trajectory_online
    boundary_constraints["orientation"]["initial"] = boundary_constraints["orientation"]["final"]
    boundary_constraints["orientation"]["final"] = rotate_x(np.pi/8)
    invariants, calculated_trajectory, calculated_movingframe, tot_time = ocp.generate_trajectory(invariant_model, boundary_constraints, step_size)

    # Print the results
    print("Invariants_start:", invariants[0])
    print("Invariants_end:", invariants[-1])
    print("Calculated Trajectory_start:", calculated_trajectory[0])
    print("Calculated Trajectory_end:", calculated_trajectory[-1])
    print("Calculated Moving Frame_start:", calculated_movingframe[0])
    print("Calculated Moving Frame_end:", calculated_movingframe[-1])
    print("Error1:", err1)
    print("Error2:", calculated_trajectory[-1].T @ boundary_constraints["orientation"]["final"])



"""
Created on Fri Sep 17 09:09:09 2021

@author: Glenn Maes

Example script of shape-preserving trajectory generation
"""


""" [Arno] This script seems incompatible with current invariants_py directory ...




import sys
sys.path.append('../implementation')
import invariants_py.calculate_invariant_trajectory as calc_traj
import tf_conversions as tf
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D

# Initialization
demo_traj_file = "sine_wave.txt"    # recorded_motion.csv
#dirname = os.path.dirname()
data_location = '../data/demonstrations/' + demo_traj_file
parameterization = 'geometric'      # {timebased, geometric}

progress_trigger_vals = [0.25, 0.5, 0.75]     # Select points along trajectory where a new trajectory needs to be generated
colors = ['b','g','r','c','m','y','k','w']
markers = ["o","s","8","p","*","h","X","d"]

startPose = tf.Frame(tf.Rotation.EulerZYX(0.0, 0.0, 0.0), tf.Vector(0.0, 0.0, 0.0))
endPose = tf.Frame(tf.Rotation.EulerZYX(0.0, 0.0, 0.0), tf.Vector(2.0, 2.0, 2.0))
# Offsets below will define how the endPose will move throughout calculation
x_offset = 0.0
y_offset = 0.5
z_offset = 0.0


#counter = 1
current_progress = 0
trajectory_list = []
invariants_list = []

# Initialize problem
traj_data = calc_traj.CalculateTrajectory(data_location, parameterization)
startPose_mat = tf.toMatrix(startPose)
endPose_mat = tf.toMatrix(endPose)

# Solve first window
traj_data.first_window(startPose_mat, endPose_mat)
trajectory_list.append(traj_data.current_pose_trajectory)
invariants_list.append(traj_data.current_invariants)

# Continued trajectory generation
for idx, progress_val in enumerate(progress_trigger_vals):
    # Determine index corresponding to start window
    N = len(trajectory_list[-1])
    new_progress = progress_val
    currentPose_index = int((new_progress-current_progress)*N)
    currentPose = trajectory_list[-1][currentPose_index]
    
    # Offset of endPose
    endPose_mat[0,3] += x_offset*idx
    endPose_mat[1,3] += y_offset*idx
    endPose_mat[2,3] += z_offset*idx
    
    # Calculation of new trajectory
    traj_data.trajectory_generation(currentPose, endPose_mat, progress_val)
    trajectory_list.append(traj_data.current_pose_trajectory)
    invariants_list.append(traj_data.current_invariants)
    
    current_progress = new_progress

# Plotting of results (translation only)
fig = plt.figure()
ax = Axes3D(fig)

for i in range(0,len(trajectory_list)):
    x = []
    y = []
    z = []
    for pose in trajectory_list[i]:
        x.append(pose[0,3])
        y.append(pose[1,3])
        z.append(pose[2,3])
    ax.scatter(x,y,z,c=colors[i],marker=markers[i],label=i)

ax.set_xlabel('x')
ax.set_ylabel('y')
ax.set_zlabel('z')

if plt.get_backend() != 'agg':
    plt.show()

"""